"""Splitting texts is commonly used as a preprocessing step before embedding and retrieving texts.

We encourage you to process your data here and define your own embedding and retrieval methods. These methods can highly depend on the product environment and may extend beyond the scope of this library.

However, the following approaches are commonly shared:

* **Document Storage:** Define how to store the documents, both raw and chunked. For example, LlamaIndex uses `Document Stores <https://docs.llamaindex.ai/en/stable/module_guides/storing/docstores/>`_ to manage ingested document chunks.

* **Document Chunking:** Segment documents into manageable chunks suitable for further processing.

* **Vectorization:** Embed each chunk and store the resulting vectors in Vector Stores. For example, LLama index utilizes `Vector Stores <https://docs.llamaindex.ai/en/stable/module_guides/storing/vector_stores/>`_.

* **Retrieval:** Leverage vectors for context retrieval.
"""

from copy import deepcopy
from typing import List, Literal, Optional
from tqdm import tqdm
import logging

from adalflow.core.component import Component
from adalflow.core.types import Document
from adalflow.core.tokenizer import Tokenizer

# TODO:
# More splitters such as PDF/JSON/HTML Splitter can be built on TextSplitter.

log = logging.getLogger(__name__)

DocumentSplitterInputType = List[Document]
DocumentSplitterOutputType = List[Document]

# customizable seperators map
SEPARATORS = {
    "page": "\f",
    "passage": "\n\n",
    "word": " ",
    "sentence": ".",
    "token": "",
}

DEFAULT_CHUNK_SIZE = 800
DEFAULT_CHUNK_OVERLAP = 200


# TODO: make it a non-component
class TextSplitter(Component):
    """
    Text Splitter for Chunking Documents

    ``TextSplitter`` first utilizes ``split_by`` to specify the text-splitting criterion and breaks the long text into smaller texts.
    Then we create a sliding window with length= ``chunk_size``. It moves at step= ``chunk_size`` - ``chunk_overlap``.
    The texts inside each window will get merged to a smaller chunk. The generated chunks from the splitted text will be returned.

    **Splitting Types**

    ``TextSplitter`` supports 2 types of splitting.

    * **Type 1:** Specify the exact text splitting point such as space<" "> and periods<".">. It is intuitive, for example, split_by "word":

    ::

        "Hello, world!" -> ["Hello, " ,"world!"]

    * **Type 2:** Use :class:`tokenizer <adalflow.core.tokenizer.Tokenizer>`. It works as:

    ::

        "Hello, world!" -> ['Hello', ',', ' world', '!']

    This aligns with how models see text in the form of tokens (`Reference <https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb>`_),
    Tokenizer reflects the real token numbers the models take in and helps the developers control budgets.

    **Definitions**

    * **split_by** specifies the split rule, i.e. the smallest unit during splitting. We support ``"word"``, ``"sentence"``, ``"page"``, ``"passage"``, and ``"token"``. The splitter utilizes the corresponding separator from the ``SEPARATORS`` dictionary.
    For Type 1 splitting, we apply ``Python str.split()`` to break the text.

    * **SEPARATORS**: Maps ``split_by`` criterions to their exact text separators, e.g., spaces <" "> for "word" or periods <"."> for "sentence".

    .. note::
        For option ``token``, its separator is "" because we directly split by a tokenizer, instead of text point.

    * **chunk_size** is the the maximum number of units in each chunk.

    * **chunk_overlap** is the number of units that each chunk should overlap. Including context at the borders prevents sudden meaning shift in text between sentences/context, especially in sentiment analysis.


    * **Splitting Details**
    Type 1:
    The ``TextSplitter`` utilizes Python's ``str.split(separator)`` method.
    Developers can refer to

    .. code-block:: none

        {
            "page": "\\f",
            "passage": "\\n",
            "word": " ",
            "sentence": "."
        }
    for exact points of text division.

    .. note::
        Developers need to determine how to assign text to each data chunk for the embedding and retrieval tasks.

    Type 2:
    We implement a tokenizer using ``cl100k_base`` encoding that aligns with how models see text in the form of tokens.
    E.g. "tiktoken is great!" -> ["t", "ik", "token", " is", " great", "!"] This helps developers control the token usage and budget better.

    * **Merge Details**
    Type 1/Type 2 create a list of split texts. ``TextSplitter`` then reattaches the specified separator to each piece of the split text, except for the last segment.
    This approach maintains the original spacing and punctuation, which is critical in contexts like natural language processing where text formatting can impact interpretations and outcomes.
    E.g. "hello world!" split by "word" will be kept as "hello " and "world!"

    * **Customization**
    You can also customize the ``SEPARATORS``. For example, by defining ``SEPARATORS`` = {"question": "?"} and setting ``split_by`` = "question", the document will be split at each ``?``, ideal for processing text structured
    as a series of questions. If you need to customize :class:`tokenizer <adalflow.core.tokenizer.Tokenizer>`, please check `Reference <https://github.com/openai/openai-cookbook/blob/main/examples/How_to_count_tokens_with_tiktoken.ipynb>`_.

    * **Integration with Other Document Types**
    This functionality is ideal for segmenting texts into sentences, words, pages, or passages, which can then be processed further for NLP applications.
    For **PDFs**, developers will need to extract the text before using the splitter. Libraries like ``PyPDF2`` or ``PDFMiner`` can be utilized for this purpose.
    ``LightRAG``'s future implementations will introduce splitters for ``JSON``, ``HTML``, ``markdown``, and ``code``.

    Example:

    .. code-block:: python

        from adalflow.components.data_process.text_splitter import TextSplitter
        from adalflow.core.types import Document

        # Configure the splitter settings
        text_splitter = TextSplitter(
            split_by="word",
            chunk_size=5,
            chunk_overlap=1
        )

        # Example document
        doc = Document(
            text="Example text. More example text. Even more text to illustrate.",
            id="doc1"
        )

        # Execute the splitting
        splitted_docs = text_splitter.call(documents=[doc])

        for doc in splitted_docs:
            print(doc)

        # Output:
        # Document(id=44a8aa37-0d16-40f0-9ca4-2e25ae5336c8, text='Example text. More example text. ', meta_data=None, vector=[], parent_doc_id=doc1, order=0, score=None)
        # Document(id=ca0af45b-4f88-49b5-97db-163da9868ea4, text='text. Even more text to ', meta_data=None, vector=[], parent_doc_id=doc1, order=1, score=None)
        # Document(id=e7b617b2-3927-4248-afce-ec0fc247ac8b, text='to illustrate.', meta_data=None, vector=[], parent_doc_id=doc1, order=2, score=None)
    """

    tokenizer = Tokenizer()

    def __init__(
        self,
        split_by: Literal["word", "sentence", "page", "passage", "token"] = "word",
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        chunk_overlap: int = DEFAULT_CHUNK_OVERLAP,
        batch_size: int = 1000,
        separators: Optional[dict] = SEPARATORS,
    ):
        """
        Initializes the TextSplitter with the specified parameters for text splitting.

        Args:
            split_by (str): The specific criterion to use for splitting the text.
                            Valid options are 'word' to split by ' ', 'sentence' to split by '.',
                            'page' to split by '\\f', 'passage' to split by '\\n\\n'.
            chunk_size (int): The size of chunks to generate after splitting. Must be greater than 0.
            chunk_overlap (int): The number of characters of overlap between chunks. Must be non-negative
                                and less than chunk_size.
            batch_size (int): The size of documents to process in each batch.
        Raises:
            ValueError: If the provided split_by is not supported, chunk_size is not greater than 0,
                        or chunk_overlap is not valid as per the given conditions.
        """
        super().__init__()

        self.split_by = split_by
        self.separators = separators
        if split_by not in self.separators:
            raise ValueError(
                f"Invalid options for split_by. You must select from {list(self.separators.keys())}."
            )

        if chunk_overlap >= chunk_size:
            raise ValueError(
                f"chunk_overlap can't be larger than or equal to chunk_size. Received chunk_size: {chunk_size}, chunk_overlap: {chunk_overlap}"
            )

        if chunk_size <= 0:
            raise ValueError(
                f"chunk_size must be greater than 0. Received value: {chunk_size}"
            )
        self.chunk_size = chunk_size

        if chunk_overlap < 0:
            raise ValueError(
                f"chunk_overlap must be non-negative. Received value: {chunk_overlap}"
            )
        self.chunk_overlap = chunk_overlap

        self.batch_size = batch_size

        log.info(
            f"Initialized TextSplitter with split_by={self.split_by}, chunk_size={self.chunk_size}, chunk_overlap={self.chunk_overlap}, batch_size={self.batch_size}"
        )

    def split_text(self, text: str) -> List[str]:
        """
        Splits the provided text into chunks.

        Splits based on the specified split_by, chunk size, and chunk overlap settings.

        Args:
            text (str): The text to split.

        Returns:
            List[str]: A list of text chunks.
        """
        log.info(
            f"Splitting text with split_by: {self.split_by}, chunk_size: {self.chunk_size}, chunk_overlap: {self.chunk_overlap}"
        )
        separator = self.separators[self.split_by]
        splits = self._split_text_into_units(text, separator)
        log.info(f"Text split into {len(splits)} parts.")
        chunks = self._merge_units_to_chunks(
            splits, self.chunk_size, self.chunk_overlap, separator
        )
        log.info(f"Text merged into {len(chunks)} chunks.")
        return chunks

    def call(self, documents: DocumentSplitterInputType) -> DocumentSplitterOutputType:
        """
        Process the splitting task on a list of documents in batch.

        Batch processes a list of documents, splitting each document's text according to the configured
        split_by, chunk size, and chunk overlap.

        Args:
            documents (List[Document]): A list of Document objects to process.

        Returns:
            List[Document]: A list of new Document objects, each containing a chunk of text from the original documents.

        Raises:
            TypeError: If 'documents' is not a list or contains non-Document objects.
            ValueError: If any document's text is None.
        """

        if not isinstance(documents, list) or any(
            not isinstance(doc, Document) for doc in documents
        ):
            log.error("Input should be a list of Documents.")
            raise TypeError("Input should be a list of Documents.")

        split_docs = []
        # Using range and batch_size to create batches
        for start_idx in tqdm(
            range(0, len(documents), self.batch_size),
            desc="Splitting Documents in Batches",
        ):
            batch_docs = documents[start_idx : start_idx + self.batch_size]

            for doc in batch_docs:
                if not isinstance(doc, Document):
                    log.error(
                        f"Each item in documents should be an instance of Document, but got {type(doc).__name__}."
                    )
                    raise TypeError(
                        f"Each item in documents should be an instance of Document, but got {type(doc).__name__}."
                    )

                if doc.text is None:
                    log.error(f"Text should not be None. Doc id: {doc.id}")
                    raise ValueError(f"Text should not be None. Doc id: {doc.id}")

                text_splits = self.split_text(doc.text)
                meta_data = deepcopy(doc.meta_data)

                split_docs.extend(
                    [
                        Document(
                            text=txt,
                            meta_data=meta_data,
                            parent_doc_id=f"{doc.id}",
                            order=i,
                            vector=[],
                        )
                        for i, txt in enumerate(text_splits)
                    ]
                )
        log.info(
            f"Processed {len(documents)} documents into {len(split_docs)} split documents."
        )
        return split_docs

    def _split_text_into_units(self, text: str, separator: str) -> List[str]:
        """Split text based on the specified separator."""
        if self.split_by == "token":
            splits = TextSplitter.tokenizer.encode(text)
        else:
            splits = text.split(separator)
        log.info(f"Text split by '{separator}' into {len(splits)} parts.")
        return splits

    def _merge_units_to_chunks(
        self, splits: List[str], chunk_size: int, chunk_overlap: int, separator: str
    ) -> List[str]:
        """
        Merge split text chunks based on the specified chunk size and overlap.
        """
        chunks = []
        # we use a window to get the text for each trunk, the window size is chunk_size, step is chunk_size - chunk_overlap
        step = chunk_size - chunk_overlap
        idx = 0

        for idx in range(0, len(splits), step):
            # 1. if the window exceeds the list of splitted string, break and process the last chunk
            # 2. if the window ends exactly the same with the splits, then break and treat the splits[idx:len(splits)] as the last chunk
            if idx + chunk_size >= len(splits):
                break
            current_splits = splits[idx : idx + chunk_size]
            # add the separator between each unit and merge the string
            # this won't be the last chunk, so we need to add the separator at the end
            if self.split_by == "token":
                chunk = current_splits  # if token, then keep the original form
            else:
                chunk = separator.join(current_splits) + separator
            chunks.append(chunk)

        if idx < len(splits):
            if self.split_by == "token":
                last_chunk = splits[idx:]  # if token, then keep the original form
            else:
                last_chunk = separator.join(
                    splits[idx:]
                )  # if not token, then join into string
            if len(last_chunk) > 0:
                chunks.append(last_chunk)

        if self.split_by == "token":
            # decode each chunk here
            chunks = [TextSplitter.tokenizer.decode(chunk) for chunk in chunks]

        log.info(f"Merged into {len(chunks)} chunks.")
        return chunks

    def _extra_repr(self) -> str:
        s = f"split_by={self.split_by}, chunk_size={self.chunk_size}, chunk_overlap={self.chunk_overlap}"
        return s



r"""Config helper functions to manage configuration and rebuilt your task pipeline.

Config format: json
(1) include attribute and entity_name to reconstruct all attributes of a pipeline.

Example:
    {  # attribute and its config to recreate the component
        "document_splitter": {
            "component_name": "DocumentSplitter",
            "component_config": {
                "split_by": "word",
                "split_length": 400,
                "split_overlap": 200,
            },
        },
        "to_embeddings": {
            "component_name": "ToEmbeddings",
            "component_config": {
                "embedder": {
                    "component_name": "Embedder",
                    "component_config": {
                        "model_client": {
                            "entity_name": "OpenAIClient",
                            "entity_config": {},
                        },
                        "model_kwargs": {
                            "model": "text-embedding-3-small",
                            "dimensions": 256,
                            "encoding_format": "float",
                        },
                    },
                    # the other config is to instantiate the entity (class and function) with the given config as arguments
                    # "entity_state": "storage/embedder.pkl", # this will load back the state of the entity
                },
                "batch_size": 100,
            },
        },
    }


(2) only include the config as arguments and does not include any entity_name or attribute.
You can use this manually to construct an entity yourself.

Example:
{
   "Embedder":   # it can be any name
    {
        "model_client": "OpenAIClient",
        "model_kwargs": {
            "api_key": "your_api_key",
            "model_name": "text-embedder"
        }
    },
    "FAISSRetriever": {
        "top_k": 2,
        "dimensions": 256,
        "vectorizer": "embedder"
    }
}
"""

from typing import Any, Dict
from adalflow.utils.registry import EntityMapping


def new_component(config: Dict[str, Any]) -> Any:
    r"""Create a single componenet from a configuration dictionary. Format type 1.

    Args:
        config (Dict[str, Any]): Configuration dictionary for the component.

    Returns:
        Any: The constructed component.
    """
    assert (
        "component_name" in config
    ), f"component_name is required in the config: {config}"

    component_name = config["component_name"]
    component_cls = EntityMapping.get(component_name)
    component_config: Dict = config.get("component_config", {})

    initialized_config = {}

    for key, value in component_config.items():
        if isinstance(value, dict) and "component_name" in value:
            # Recursively construct sub-entities
            initialized_config[key] = new_component(value)
        else:
            initialized_config[key] = value

    return component_cls(**initialized_config)


def new_components_from_config(config: Dict[str, Dict[str, Any]]) -> Dict[str, Any]:
    r"""Construct multiple components from a configuration dictionary. Format type 1"""
    components = {}
    for attr, component_config in config.items():
        components[attr] = new_component(component_config)
    return components



from typing import Dict, Type


class EntityMapping:
    __doc__ = r"""A registry for entities, components,classes, function.

    This can be used to configure classes, functions, or components in a registry.
    """
    _registry: Dict[str, Type] = {}

    @classmethod
    def register(cls, name: str, entity_cls: Type):
        cls._registry[name] = entity_cls

    @classmethod
    def get(cls, name: str) -> Type:
        return cls._registry[name] if name in cls._registry else None

    @classmethod
    def get_all(cls) -> Dict[str, Type]:
        return cls._registry



"""This is the default system prompt template used in the AdalFlow.

Use :ref:`Prompt <core-prompt_builder>` class  to manage it.
"""

__all__ = [
    "ADALFLOW_DEFAULT_PROMPT_ARGS",
    "ADALFLOW_DEFAULT_PROMPT_TRAINABLE_PARAMS",
    "SIMPLE_DEFAULT_ADALFLOW_SYSTEM_PROMPT",
    "DEFAULT_ADALFLOW_SYSTEM_PROMPT",
]
# TODO: potentially make a data class for this
ADALFLOW_DEFAULT_PROMPT_ARGS = [
    "task_desc_str",  # task description
    "output_format_str",  # output format of the task
    "tools_str",  # tools used in the task
    "examples_str",  # examples of the task
    "chat_history_str",  # chat history of the user
    "context_str",  # context of the user query
    "steps_str",  # used in agent steps
    "input_str",  # user query or input
]

ADALFLOW_DEFAULT_PROMPT_TRAINABLE_PARAMS = [
    "task_desc_str",
    # "output_format_str",
    "examples_str",
]

SIMPLE_DEFAULT_ADALFLOW_SYSTEM_PROMPT = r"""<SYS>{{task_desc_str}}</SYS>
User: {{input_str}}
You:"""

DEFAULT_ADALFLOW_SYSTEM_PROMPT = r"""<START_OF_SYSTEM_PROMPT>
{# task desc #}
{% if task_desc_str %}
{{task_desc_str}}
{% else %}
You are a helpful assistant.
{% endif %}
{#input format#}
{% if input_format_str %}
<INPUT_FORMAT>
{{input_format_str}}
</INPUT_FORMAT>
{% endif %}
{# output format #}
{% if output_format_str %}
<OUTPUT_FORMAT>
{{output_format_str}}
</OUTPUT_FORMAT>
{% endif %}
{# tools #}
{% if tools_str %}
<TOOLS>
{{tools_str}}
</TOOLS>
{% endif %}
{# example #}
{% if examples_str %}
<EXAMPLES>
{{examples_str}}
</EXAMPLES>
{% endif %}
{# chat history #}
{% if chat_history_str %}
<CHAT_HISTORY>
{{chat_history_str}}
</CHAT_HISTORY>
{% endif %}
{#contex#}
{% if context_str %}
<CONTEXT>
{{context_str}}
</CONTEXT>
{% endif %}
<END_OF_SYSTEM_PROMPT>
<START_OF_USER_PROMPT>
{% if input_str %}
{{input_str}}
{% endif %}
<END_OF_USER_PROMPT>
{# steps #}
{% if steps_str %}
<START_OF_ASSISTANT_STEPS>
{{steps_str}}
<END_OF_ASSISTANT_STEPS>
{% endif %}
"""
"""This is the default system prompt template used in the AdalFlow.

Use :ref:`Prompt <core-prompt_builder>` class  to manage it.
"""



"""The most commonly used output parsers for the Generator.

Includes:
- YamlOutputParser: YAML output parser using dataclass for schema extraction.
- JsonOutputParser: JSON output parser using dataclass for schema extraction.
- ListOutputParser: List output parser to parse list of objects from the string.
- BooleanOutputParser: Boolean output parser to parse boolean values from the string.
"""

from dataclasses import is_dataclass
from typing import Dict, Any, Optional, List
import logging

from adalflow.core.component import DataComponent
from adalflow.core.prompt_builder import Prompt
from adalflow.core.string_parser import YamlParser, ListParser, JsonParser
from adalflow.core.base_data_class import DataClass, DataClassFormatType
from adalflow.core.base_data_class import ExcludeType, IncludeType


__all__ = [
    "OutputParser",
    "YamlOutputParser",
    "JsonOutputParser",
    "ListOutputParser",
    "BooleanOutputParser",
]

log = logging.getLogger(__name__)

# TODO: delete examples here
JSON_OUTPUT_FORMAT = r"""Your output should be formatted as a standard JSON instance with the following schema:
```
{{schema}}
```
{% if example %}
Examples:
```
{{example}}
```
{% endif %}
-Make sure to always enclose the JSON output in triple backticks (```). Please do not add anything other than valid JSON output!
-Use double quotes for the keys and string values.
-DO NOT mistaken the "properties" and "type" in the schema as the actual fields in the JSON output.
-Follow the JSON formatting conventions."""

YAML_OUTPUT_FORMAT = r"""Your output should be formatted as a standard YAML instance with the following schema:
```
{{schema}}
```
{% if example %}
Examples:
```
{{example}}
```
{% endif %}

-Make sure to always enclose the YAML output in triple backticks (```). Please do not add anything other than valid YAML output!
-Follow the YAML formatting conventions with an indent of 2 spaces.
-DO NOT mistaken the "properties" and "type" in the schema as the actual fields in the YAML output.
-Quote the string values properly."""

LIST_OUTPUT_FORMAT = r"""Your output should be formatted as a standard Python list.
- Start the list with '[' and end with ']'
- DO NOT mistaken the "properties" and "type" in the schema as the actual fields in the list output.
"""


YAML_OUTPUT_PARSER_OUTPUT_TYPE = Dict[str, Any]


class OutputParser(DataComponent):
    __doc__ = r"""The abstract class for all output parsers.

    On top of the basic string Parser, it handles structured data interaction:
    1. format_instructions: Return the formatted instructions to use in prompt for the output format.
    2. call: Parse the output string to the desired format and return the parsed output via yaml or json.

    This interface helps users customize output parsers with consistent interfaces for the Generator.
    Even though you don't always need to subclass it.

    AdalFlow uses two core classes:
    1. the Prompt to format output instruction
    2. A string parser from core.string_parser for response parsing.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__()
        pass

    def format_instructions(self) -> str:
        r"""Return the formatted instructions to use in prompt for the output format."""
        raise NotImplementedError("This is an abstract method.")

    def call(self, input: str) -> Any:
        r"""Parse the output string to the desired format and return the parsed output."""
        raise NotImplementedError("This is an abstract method.")


class YamlOutputParser(OutputParser):
    __doc__ = r"""YAML output parser using dataclass for schema extraction.

    .. note::
        Only use yaml for simple dataclass objects. For complex objects, use JSON.

    Args:
        data_class (Type): The dataclass to extract the schema for the YAML output.
        example (Type, optional): The example dataclass object to show in the prompt. Defaults to None.
        yaml_output_format_template (str, optional): The template for the YAML output format. Defaults to YAML_OUTPUT_FORMAT.
        output_processors (Component, optional): The output processors to parse the YAML string to JSON object. Defaults to YamlParser().

    Examples:

    >>> from prompts.outputs import YamlOutputParser
    >>> from dataclasses import dataclass, field
    >>> from typing import List
    >>>
    >>> @dataclass
    >>> class ThoughtAction:
    >>>     thought: str = field(metadata={"description": "Reasoning behind the answer"}) # required field
    >>>     answer: str = field(metadata={"description": "Your answer to the question"}, default=None) # optional field
    >>>
    >>> # If you want to parse it back to the dataclass, you can add a from_dict method to the dataclass
    >>> # def from_dict(self, d: Dict[str, Any]) -> "ThoughtAction":
    >>> #     return ThoughtAction(**d)
    >>>
    >>> yaml_parser = YamlOutputParser(data_class_for_yaml=ThoughtAction)
    >>> yaml_format_instructions = yaml_parser.format_instructions()
    >>> print(yaml_format_instructions)
    >>> yaml_str = '''The output should be formatted as a standard YAML instance with the following JSON schema:
    >>> ```
    >>> 'thought': {'type': 'str', 'description': 'Reasoning behind the answer', 'required': True}, 'answer': {'type': 'str', 'description': '
    >>> Your answer to the question', 'required': False, 'default': None}
    >>> ```
    >>> -Make sure to always enclose the YAML output in triple backticks (```). Please do not add anything other than valid YAML output!
    >>> -Follow the YAML formatting conventions with an indent of 2 spaces.
    >>> '''
    >>> # use it in the generator
    >>> task_desc_str = "You are a helpful assistant who answers user query. "+yaml_format_instructions
    >>> generator = Generator(output_processors=yaml_parser, ..., preset_prompt_kwargs={"task_desc_str": task_desc_str})
    >>> generator("Should i be a doctor?")
    """

    def __init__(
        self,
        data_class: DataClass,
        examples: List[DataClass] = None,
        include_fields: IncludeType = None,
        exclude_fields: ExcludeType = None,
        return_data_class: bool = False,
    ):

        super().__init__()
        if not is_dataclass(data_class):
            raise TypeError(f"Provided class is not a dataclass: {data_class}")

        if not issubclass(data_class, DataClass):
            raise TypeError(
                f"Provided class is not a subclass of DataClass: {data_class}"
            )

        # ensure example is instance of data class and initiated
        if examples is not None and not isinstance(examples[0], data_class):
            raise TypeError(
                f"Provided example is not an instance of the data class: {data_class}"
            )
        self._return_data_class = return_data_class
        self._exclude_fields = exclude_fields
        self._include_fields = include_fields
        self.data_class: DataClass = data_class
        self.output_format_prompt = Prompt(template=YAML_OUTPUT_FORMAT)
        self.output_processors = YamlParser()
        self.examples = examples

    def format_instructions(
        self,
        format_type: Optional[DataClassFormatType] = None,
    ) -> str:
        r"""Return the formatted instructions to use in prompt for the YAML output format.

        Args:
            format_type (DataClassFormatType, optional): The format type to show in the prompt.
                Defaults to DataClassFormatType.SIGNATURE_YAML for less token usage.
                Options: DataClassFormatType.SIGNATURE_YAML, DataClassFormatType.SIGNATURE_JSON, DataClassFormatType.SCHEMA.
            exclude (List[str], optional): The fields to exclude from the schema of the data class.
        """
        format_type = format_type or DataClassFormatType.SIGNATURE_YAML
        schema = self.data_class.format_class_str(
            format_type=format_type,
            exclude=self._exclude_fields,
            include=self._include_fields,
        )
        # convert example to string, convert data class to yaml string
        example_str = ""
        try:
            if self.examples and len(self.examples) > 0:
                for example in self.examples:
                    per_example_str = example.format_example_str(
                        format_type=DataClassFormatType.EXAMPLE_YAML,
                        exclude=self._exclude_fields,
                        include=self._include_fields,
                    )
                    example_str += f"{per_example_str}\n________\n"
                # remove the last new line
                example_str = example_str[:-1]
                log.debug(f"{__class__.__name__} example_str: {example_str}")

        except Exception as e:
            log.error(f"Error in formatting example for {__class__.__name__}, {e}")
            example_str = None

        return self.output_format_prompt(schema=schema, example=example_str)

    def call(self, input: str) -> YAML_OUTPUT_PARSER_OUTPUT_TYPE:
        r"""Parse the YAML string to JSON object and return the JSON object."""
        try:
            output_dict = self.output_processors(input)
            if self._return_data_class:
                return self.data_class.from_dict(output_dict)
            return output_dict
        except Exception as e:
            log.error(f"Error in parsing YAML to JSON: {e}")
            raise e

    def _extra_repr(self) -> str:
        s = f"data_class={self.data_class.__name__}, examples={self.examples}, exclude_fields={self._exclude_fields}, \
        include_fields={self._include_fields},\return_data_class={self._return_data_class}"
        return s


class JsonOutputParser(OutputParser):
    def __init__(
        self,
        data_class: DataClass,
        examples: List[DataClass] = None,
        include_fields: IncludeType = None,
        exclude_fields: ExcludeType = None,
        return_data_class: bool = False,
    ):
        super().__init__()
        if not is_dataclass(data_class):
            raise TypeError(f"Provided class is not a dataclass: {data_class}")

        if not issubclass(data_class, DataClass):
            raise TypeError(
                f"Provided class is not a subclass of DataClass: {data_class}"
            )

        if examples is not None and not isinstance(examples[0], data_class):
            raise TypeError(
                f"Provided example is not an instance of the data class: {data_class}"
            )
        self._return_data_class = return_data_class
        self._exclude_fields = exclude_fields
        self._include_fields = include_fields
        template = JSON_OUTPUT_FORMAT
        self.data_class: DataClass = data_class
        self.output_format_prompt = Prompt(template=template)
        self.output_processors = JsonParser()
        self.examples = examples

    def format_instructions(
        self,
        format_type: Optional[DataClassFormatType] = None,
    ) -> str:
        r"""Return the formatted instructions to use in prompt for the JSON output format.

        Args:
            format_type (DataClassFormatType, optional): The format type to show in the prompt.
                Defaults to DataClassFormatType.SIGNATURE_JSON for less token usage compared with DataClassFormatType.SCHEMA.
                Options: DataClassFormatType.SIGNATURE_YAML, DataClassFormatType.SIGNATURE_JSON, DataClassFormatType.SCHEMA.
        """
        format_type = format_type or DataClassFormatType.SIGNATURE_JSON
        schema = self.data_class.format_class_str(
            format_type=format_type,
            exclude=self._exclude_fields,
            include=self._include_fields,
        )
        example_str = ""
        try:
            if self.examples and len(self.examples) > 0:
                for example in self.examples:
                    per_example_str = example.format_example_str(
                        format_type=DataClassFormatType.EXAMPLE_JSON,
                        exclude=self._exclude_fields,
                        include=self._include_fields,
                    )
                    example_str += f"{per_example_str}\n________\n"
                # remove the last new line
                example_str = example_str[:-1]
                log.debug(f"{__class__.__name__} example_str: {example_str}")

        except Exception as e:
            log.error(f"Error in formatting example for {__class__.__name__}, {e}")
            example_str = None
        return self.output_format_prompt(schema=schema, example=example_str)

    def call(self, input: str) -> Any:
        try:
            output_dict = self.output_processors(input)
            log.debug(f"{__class__.__name__} output_dict: {output_dict}")

        except Exception as e:
            log.error(f"Error in parsing JSON to JSON: {e}")
            raise e
        try:
            if self._return_data_class:
                return self.data_class.from_dict(output_dict)
            return output_dict
        except Exception as e:
            log.error(f"Error in converting dict to data class: {e}")
            raise e

    def _extra_repr(self) -> str:
        s = f"""data_class={self.data_class.__name__}, examples={self.examples}, exclude_fields={self._exclude_fields}, \
            include_fields={self._include_fields}, return_data_class={self._return_data_class}"""
        return s


class ListOutputParser(OutputParser):
    __doc__ = r"""List output parser to parse list of objects from the string."""

    def __init__(self, list_output_format_template: str = LIST_OUTPUT_FORMAT):
        super().__init__()
        self.list_output_format_prompt = Prompt(template=list_output_format_template)
        self.output_processors = ListParser()

    def format_instructions(self) -> str:
        return self.list_output_format_prompt()

    def call(self, input: str) -> list:
        return self.output_processors(input)


def _parse_boolean_from_str(input: str) -> Optional[bool]:
    input = input.strip()
    if "true" in input.lower():
        return True
    elif "false" in input.lower():
        return False
    else:
        return None


class BooleanOutputParser(OutputParser):
    __doc__ = r"""Boolean output parser to parse boolean values from the string."""

    def __init__(self):
        super().__init__()
        self.output_processors = None

    def format_instructions(self) -> str:
        return "The output should be a boolean value. True or False."

    def call(self, input: str) -> bool:

        input = input.strip()
        output = None
        # evaluate the expression to get the boolean value
        try:
            output = eval(input)
            if isinstance(output, bool):
                return output
            # go to string parsing
            output = _parse_boolean_from_str(input)
            return output
        except Exception as e:
            # try to do regex matching for boolean values
            log.info(f"Error: {e}")
            output = _parse_boolean_from_str(input)
            return output



r"""Based and optimized from ORPO llm optimizer.

https://arxiv.org/abs/2309.03409
Source code: https://github.com/google-deepmind/opro
"""

from typing import Dict, Any, List, Optional, TYPE_CHECKING
from dataclasses import field, dataclass

from adalflow.core.base_data_class import DataClass

if TYPE_CHECKING:

    from adalflow.core.model_client import ModelClient


from adalflow.optim.optimizer import Optimizer, ParamsT

# TODO: add the responses and gts
LLM_OPTIMIZER_TEMPLATE = r"""<SYS>
Your task is to generate an new instruction that can score higher than all previous instructions.
{# starter instruction #}
{%if starter_instruction %}
<STARTER_INS>
Here is the starter instruction:
{{starter_instruction}}
</STARTER_INS>
{% endif %}

{# Current instruction #}
{% if current_instruction %}
<CURRENT_INS>
Here is the current instruction:
{{current_instruction}}
</CURRENT_INS>
{% endif %}

{# Training samples on current instruction #}
{% if training_samples %}
<TRAINING_SAMPLES>
Here are some training samples on the current instruction:
{% for sample in training_samples %}
- {{sample}}
{% endfor %}
</TRAINING_SAMPLES>
{% endif %}

{# history #}
{% if instructions %}
<HISTORY_INS>
Below are some of your previous instructions and their scores, the higher the score the better the instruction:
{% for instruction in instructions %}
- {{loop.index}}.
- text: {{instruction.text}}
- score: {{instruction.score}})
{% if instruction.responses is defined %}
- responses: {{instruction.responses}}
{% endif %}
{% if instruction.gts is defined %}
- gts: {{instruction.gts}}
{% endif %}
____
{% endfor %}
</HISTORY_INS>
{% endif %}

{# More task specification #}
- Your new instruction should be different from all previous instructions.
- It should be clear, concise, and effective.
- Do not change core information provided in the starter instruction.
</SYS>
"""


@dataclass
class Instruction(DataClass):
    # prefix will be the same as text
    text: str = field(metadata={"desc": "The instruction text"})
    score: float = field(
        metadata={"desc": "The score of the instruction, range from 0 to 1"}
    )
    responses: Optional[List[str]] = field(
        metadata={"desc": "The responses of using the instruction"}, default=None
    )
    gts: Optional[List[str]] = field(
        metadata={"desc": "The ground truth of the task"}, default=None
    )
    # for classification, use the named labels instead of label index


# TODO: adapt this to multiple parameters who requires_opt
# TODO: combine batch errors few shot
# TODO: combine few-shot with llm optimizer
# TODO: support multi processors and multiple proposals at the same time
class LLMOptimizer(Optimizer):
    __doc__ = r"""Default LLM optimizer for task instruction.

    User should always provide a starter instruction which provides the source of truth for the task.
    The optimizer will generate new instructions that can score higher than all previous instructions.
    """

    def __init__(
        self,
        params: ParamsT,
        model_client: "ModelClient",
        model_kwargs: Dict[str, Any],
    ):
        r"""Initialize the generator with the model client and the model kwargs."""
        super().__init__()
        from adalflow.core.generator import Generator

        self.instruction_parameter = list(params)[
            0
        ]  # for now only support one parameter
        # Ensure the temperature is at least 1
        model_kwargs["temperature"] = max(1, model_kwargs.get("temperature", 1))

        self.instruction_history: List[Instruction] = (
            []
        )  # trace the history of the instructions
        self.starter_instruction: Optional[str] = None
        if self.instruction_parameter.data is not None:
            self.starter_instruction = self.instruction_parameter.data
        # self.proposed: Optional[str] = None
        # self.current: Optional[str] = None
        self.prompt_kwargs = {
            "starter_instruction": self.starter_instruction,
            "instructions": self.instruction_history,
        }
        self.generator = Generator(
            model_client=model_client,
            model_kwargs=model_kwargs,
            template=LLM_OPTIMIZER_TEMPLATE,
            prompt_kwargs=self.prompt_kwargs,
        )
        self.proposing = False

    def reset(self):
        self.instruction_history = []

    def zero_grad(self):
        pass

    def propose(self, training_samples: List[str] = None):
        r"""Propose a new instruction using the generator."""
        from adalflow.core.generator import GeneratorOutput

        if self.proposing:
            raise ValueError("Already proposing a new instruction.")
        max_run: int = 5
        current_instruction = self.instruction_parameter.data
        additional_prompt_kwargs = {
            "current_instruction": current_instruction,
            "training_samples": training_samples,
        }
        for _ in range(max_run):
            instruction: GeneratorOutput = self.generator(
                prompt_kwargs=additional_prompt_kwargs
            )
            self.generator.print_prompt(**additional_prompt_kwargs)
            if instruction.data is not None:
                instruction = instruction.data
                break

        # self.proposed = instruction
        self.instruction_parameter.propose_data(instruction)
        self.proposing = True

    # Step
    def step(self, score: float):
        r"""Load the proposed instruction to the current instruction to complete the optimization step."""
        if not self.proposing:
            raise ValueError("No proposed instruction to step.")

        self.instruction_parameter.step_data()
        data = self.instruction_parameter.data
        self.instruction_history.append(Instruction(text=data, score=score))
        self.proposing = False

    # revert
    def revert(self):
        r"""The proposed is not leading to a better instruction, reset the parameter."""
        if not self.proposing:
            raise ValueError("No proposed instruction to revert.")
        # self.proposed = None
        self.instruction_parameter.revert_data()
        self.proposing = False



r"""The base class for all retrievers who in particular retrieve documents from a given database."""

from typing import List, Optional, Generic, Any, Callable, Union
import logging

from adalflow.core.types import (
    RetrieverQueriesType,
    RetrieverQueryType,
    RetrieverDocumentType,
    RetrieverDocumentsType,
    RetrieverOutputType,
)
from adalflow.optim.grad_component import GradComponent

from adalflow.optim.parameter import Parameter, OutputParameter
from adalflow.optim.types import ParameterType

log = logging.getLogger(__name__)


# TODO: tracing retriever in the diagnose files using callback manager
class Retriever(GradComponent, Generic[RetrieverDocumentType, RetrieverQueryType]):
    __doc__ = r"""The base class for all retrievers.

    Retriever will manage its own index and retrieve in format of RetrieverOutput

    Args:
        indexed (bool, optional): whether the retriever has an index. Defaults to False.
        index_keys (List[str], optional): attributes that define the index that can be used to restore the retriever. Defaults to [].

    The key method :meth:`build_index_from_documents` is the method to build the index from the documents.
    ``documents`` is a sequence of any type of document. With ``document_map_func``, you can map the document
    of Any type to the specific type ``RetrieverDocumentType`` that the retriever expects.

    note:
    To get the state of the retriever, leverage the :methd: "from_dict" and "to_dict" methods of the base class Component.


    """

    indexed: bool = False
    index_keys: List[str] = []  # attributes that define the index
    name: str = "Retriever"
    top_k: int

    def __init__(self, *args, **kwargs):
        super().__init__(desc="Retrieve a list of documents using a query", **kwargs)

    def reset_index(self):
        r"""Initialize/reset any attributes/states for the index."""
        raise NotImplementedError("reset_index is not implemented")

    def build_index_from_documents(
        self,
        documents: RetrieverDocumentsType,
        document_map_func: Optional[Callable[[Any], RetrieverDocumentType]] = None,
        **kwargs,
    ):
        r"""Built index from the [document_map_func(doc) for doc in documents]."""
        raise NotImplementedError(
            "build_index_from_documents and input_field_map_func is not implemented"
        )

    def save_to_file(self, path: str):
        r"""Save the state, including the index to a file.

        Optional for subclass to implement a default persistence method.
        Subclass can leverge component's `to_dict` method to get the states and choose to save them in any file format.
        """
        pass

    @classmethod
    def load_from_file(cls, path: str):
        r"""Load the state, including index from a file to restore the retriever.

        Subclass can leverge component's `from_dict` method to restore the states from the file.
        """
        pass

    def call(
        self,
        input: RetrieverQueriesType,
        top_k: Optional[int] = None,
        id: str = None,  # for tracing, diagnosing, and training
        **kwargs,
    ) -> RetrieverOutputType:
        raise NotImplementedError("retrieve is not implemented")

    async def acall(
        self,
        input: RetrieverQueriesType,
        top_k: Optional[int] = None,
        id: str = None,  # for tracing, diagnosing, and training
        **kwargs,
    ) -> RetrieverOutputType:
        raise NotImplementedError("Async retrieve is not implemented")

    def forward(
        self,
        input: Union[RetrieverQueriesType, Parameter],
        top_k: Optional[
            int
        ] = None,  # TODO: top_k can be trained in the future if its formulated as a parameter
        id: str = None,  # for tracing, diagnosing, and training
        **kwargs,
    ) -> Parameter:
        r"""Customized forward on top of the GradComponent forward method.

        To track the input as Parameters and set the parameter type as RETRIEVER_OUTPUT in the response.
        """
        # convert input to parameter if it is not
        if not isinstance(input, Parameter):
            input = Parameter(
                data=input,
                name="input",
                requires_opt=True,
                param_type=ParameterType.INPUT,
            )
        # trace the top_k in the DAG graph
        top_k = Parameter(
            data=top_k or self.top_k,
            name="top_k",
            requires_opt=False,
            param_type=ParameterType.HYPERPARAM,
        )
        if input is None:
            raise ValueError("Input cannot be empty")
        response: OutputParameter = super().forward(input, top_k=top_k, id=id, **kwargs)
        if not isinstance(response, OutputParameter):
            raise ValueError(
                f"Retriever forward: Expect OutputParameter, but got {type(response)}"
            )
        response.trace_forward_pass(
            input_args={"input": input, "top_k": top_k},
            full_response=response.data,
            id=self.id,
            name=self.name,
        )
        response.param_type = (
            ParameterType.RETRIEVER_OUTPUT
        )  # be more specific about the type
        return response



"""Base Classes for AdalFlow Optimizers, including Optimizer, TextOptimizer, and DemoOptimizer."""

from typing import Dict, Any, Union, Iterable, Sequence, List

from adalflow.optim.parameter import Parameter
from adalflow.core.base_data_class import DataClass


ParamsT = Union[Iterable[Parameter], Iterable[Dict[str, Any]]]


class Optimizer:
    __doc__ = r"""Base class for all optimizers."""

    proposing: bool = False
    params: ParamsT

    def state_dict(self):
        pass

    def propose(self, *args, **kwargs):
        raise NotImplementedError("propose method is not implemented")

    def step(self, *args, **kwargs):
        raise NotImplementedError("step method is not implemented")

    def revert(self, *args, **kwargs):
        raise NotImplementedError("revert method is not implemented")


class TextOptimizer(Optimizer):
    __doc__ = r"""Base class for all text optimizers.

    Text optimizer is via textual gradient descent, which is a variant of gradient descent that optimizes the text directly.
    It will generate new values for a given text prompt.This includes:
    - System prompt
    - output format
    - prompt template
    """

    def __init__(self, *args, **kwargs):
        pass

    def zero_grad(self):
        """Clear all the gradients of the parameters."""
        raise NotImplementedError("zero_grad method is not implemented")


class DemoOptimizer(Optimizer):

    __doc__ = r"""Base class for all demo optimizers.

    Demo optimizer are few-shot optimization, where it will sample raw examples from train dataset or bootstrap examples from the model's output.
    It will work with a sampler to generate new values for a given text prompt.

    If bootstrap is used, it will require a teacher genearator to generate the examples.
    """

    _traces: Dict[str, Any]  # key: parameter_id (demo)
    dataset: Sequence[DataClass]
    _weighted: bool
    exclude_input_fields_from_bootstrap_demos: bool = False

    def __init__(
        self,
        weighted: bool = True,
        dataset: Sequence[DataClass] = None,
        exclude_input_fields_from_bootstrap_demos: bool = False,
        *args,
        **kwargs,
    ):
        self._weighted = weighted
        self.dataset = dataset
        self.exclude_input_fields_from_bootstrap_demos = (
            exclude_input_fields_from_bootstrap_demos
        )

    def use_weighted_sampling(self, weighted: bool):
        self._weighted = weighted

    def config_shots(self, *args, **kwargs):
        r"""Initialize the samples for each parameter."""
        raise NotImplementedError("init method is not implemented")

    def set_dataset(self, dataset: Sequence[DataClass]):
        r"""Set the dataset for the optimizer."""
        self.dataset = dataset

    def add_scores(self, ids: List[str], scores: List[float], *args, **kwargs):
        r"""Add scores to the optimizer."""
        raise NotImplementedError("add_scores method is not implemented")



# from hotpotqa github
import re

import string
from collections import Counter


def normalize_answer(s):

    def remove_articles(text):
        return re.sub(r"\b(a|an|the)\b", " ", text)

    def white_space_fix(text):
        return " ".join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def f1_score(y: str, y_gt: str) -> float:
    if not isinstance(y, str) or not isinstance(y_gt, str):
        raise ValueError(f"y: {y},{type(y)}, y_gt: {y_gt},{type(y_gt)} must be string.")
    prediction_tokens = normalize_answer(y).split()
    ground_truth_tokens = normalize_answer(y_gt).split()

    common = Counter(prediction_tokens) & Counter(ground_truth_tokens)
    num_same = sum(common.values())

    if len(prediction_tokens) == len(ground_truth_tokens) == 0:
        # Unlike most tasks, QReCC and SQuAD-2.0 assign 1.0 in this edge case. We don't for uniformity.
        print(
            "\n#> F1 Metric: Rare edge case of len(prediction_tokens) == len(ground_truth_tokens) == 0.\n"
        )

    if num_same == 0:
        return 0

    precision = 1.0 * num_same / len(prediction_tokens)
    recall = 1.0 * num_same / len(ground_truth_tokens)
    f1 = (2 * precision * recall) / (precision + recall)

    return f1



"""Adapted and optimized boostrap fewshot optimizer:

DSPy: Com-piling declarative language model calls into state-of-the-art pipelines."""

from typing import List, Optional, Dict
import logging

from adalflow.core.base_data_class import DataClass


from adalflow.optim.parameter import Parameter


from adalflow.core.functional import random_sample
from adalflow.optim.optimizer import DemoOptimizer
from adalflow.optim.types import ParameterType
from adalflow.utils import printc

log = logging.getLogger(__name__)


class BootstrapFewShot(DemoOptimizer):
    __doc__ = r"""BootstrapFewShot performs few-shot sampling used in few-shot ICL.

    It will be used to optimize paramters of demos.
    Based on research from AdalFlow team and DsPy library.

    Compared with Dspy's version:
     1. we added weighted sampling for both the raw and augmented demos
        to prioritize failed demos but successful in augmented demos based on the evaluation score
        while we backpropagate the demo samples.
     2. In default, we exclude the input fields from the augmented demos. Our reserch finds that
        using the reasoning demostrations from teacher model can be more effective in some cases than taking both inputs and output
        samples and be more token efficient.

    Reference:
    - DsPy: Com-piling declarative language model calls into state-of-the-art pipelines.
    """
    exclude_input_fields_from_bootstrap_demos: bool

    def __init__(
        self,
        params: List[Parameter],
        raw_shots: Optional[int] = None,
        bootstrap_shots: Optional[int] = None,
        dataset: Optional[List[DataClass]] = None,
        weighted: bool = True,
        exclude_input_fields_from_bootstrap_demos: bool = False,
    ):
        super().__init__(weighted=weighted, dataset=dataset)
        self.params = [
            param
            for param in params
            if param.requires_opt and param.param_type == ParameterType.DEMOS
        ]
        log.info(f"BootstrapFewShot: {[p.name for p in self.params]}")

        print(f"BootstrapFewShot: {[p.name for p in self.params]}")

        self._raw_shots = raw_shots
        self._bootstrap_shots = bootstrap_shots

        self.proposing = False
        self._teacher_scores: Dict[str, float] = {}  # data id to score
        self._student_scores: Dict[str, float] = {}  # data id to score
        self.exclude_input_fields_from_bootstrap_demos = (
            exclude_input_fields_from_bootstrap_demos
        )

    # TODO: use the scores from the backward engine (optionally) on the demo parameters
    # needs to make a decision on which
    # this score does not make sense for multiple demo parameters
    def add_scores(self, ids: List[str], scores: List[float], is_teacher: bool = True):
        r"""Add scores for each demo via _teacher_scores or _student_scores."""
        if len(ids) != len(scores):
            raise ValueError(
                f"ids and scores must have the same length, got ids: {ids}, scores: {scores}"
            )

        for score in scores:

            if not isinstance(score, float):
                raise ValueError(
                    f"score must be a float, got {type(score)}, score: {score}"
                )
            if score < 0 or score > 1:
                raise ValueError("score must be in range [0, 1]")

        target = self._teacher_scores if is_teacher else self._student_scores
        for i, id in enumerate(ids):
            target[id] = scores[i]

    def config_shots(self, raw_shots: int, bootstrap_shots: int):
        self._raw_shots = raw_shots
        self._bootstrap_shots = bootstrap_shots

    def config_dataset(self, dataset: List[DataClass]):
        self.dataset = dataset

    def _pre_check(self):
        if not self.dataset:
            raise ValueError("dataset must be provided")
        if self._raw_shots is None or self._bootstrap_shots is None:
            raise ValueError("raw_shots and bootstrap_shots must be provided")

    @property
    def num_shots(self) -> int:
        return (
            self._raw_shots + self._bootstrap_shots
            if self._raw_shots and self._bootstrap_shots
            else 0
        )

    def sample(
        self,
        augmented_demos: Dict[str, DataClass],
        demos: Dict[str, DataClass],
        dataset: List[DataClass],
        raw_shots: int,
        bootstrap_shots: int,
        weighted: bool = True,
    ):
        r"""Performs weighted sampling, ensure the score is in range [0, 1]. The higher score means better accuracy."""
        # 1. sample from augmented demos (from teacher)
        # set weights to be score
        # add 1 to all score to avoid negative weights
        augmented_options = list(augmented_demos.values())

        # get the teacher scores length and the augmented demos length
        len_teacher_scores = len(self._teacher_scores)
        len_augmented_options = len(augmented_options)
        print(f"len_teacher_scores: {len_teacher_scores}")
        print(f"len_augmented_options: {len_augmented_options}")
        weights = None
        if weighted:
            weights: List[float] = []
            for demo in augmented_options:
                demo_score = self._teacher_scores.get(demo.id, None)
                if demo_score is None:
                    raise ValueError(
                        f"score must be provided for each demo, id: {demo.id}, all scores: {self._teacher_scores}"
                    )
                if demo_score < 0 or demo_score > 1:
                    raise ValueError(f"score must be in range [0, 1], got {demo_score}")

                w = demo_score
                student_demo_score = self._student_scores.get(demo.id, None)

                if student_demo_score is not None:
                    if student_demo_score < 0 or student_demo_score > 1:
                        raise ValueError(
                            f"score must be in range [0, 1], got {student_demo_score}"
                        )
                    # if demo.id in demos and demos[demo.id].score is not None:
                    w = (
                        w
                        - student_demo_score
                        # w - demos[demo.id].score
                    )  # assign higher weights to failed demos but successful in augmented
                    if w < 0:
                        w = 0
                weights.append(w)

        # print(f"augs: {augmented_options}")
        sampled_augmented_demos = (
            random_sample(
                augmented_options, bootstrap_shots, replace=False, weights=weights
            )
            if len(augmented_options) > 0
            else []
        )

        # 2. sample from raw demos
        # exclude the sampled augmented demos
        # TODO: ensure all data points has unique ids
        filtered_dataset = list(
            filter(
                lambda x: x.id
                not in set([demo.id for demo in sampled_augmented_demos]),
                dataset,
            )
        )
        if len(filtered_dataset) == 0:
            # If no demos left we will get raw_weights [], sum to 0
            return sampled_augmented_demos, []
        # assigne weights 0 to all options
        raw_weights = None
        if weighted:
            raw_weights = [0.0] * len(filtered_dataset)
            # for those exist in the demos, assign higher score with failed demos
            for i, demo in enumerate(filtered_dataset):
                student_demo_score = self._student_scores.get(demo.id, None)

                if student_demo_score is not None:
                    # ensure the score is in range [0, 1]
                    if student_demo_score < 0 or student_demo_score > 1:
                        raise ValueError(
                            f"score must be in range [0, 1], got {student_demo_score}"
                        )
                    raw_weights[i] = 1 - student_demo_score

        sampled_raw_demos = random_sample(
            filtered_dataset, raw_shots, replace=False, weights=raw_weights
        )
        return sampled_augmented_demos, sampled_raw_demos

    @staticmethod
    def samples_to_str(
        samples: List[DataClass], augmented: bool = False, exclude_inputs: bool = False
    ) -> str:
        sample_strs = []
        for sample in samples:
            try:

                # process the input fields
                if augmented:
                    exclude_fields = ["id", "score"]
                    if exclude_inputs:
                        exclude_fields.extend(sample.get_input_fields())
                    yaml_str = sample.to_yaml(exclude=exclude_fields)

                else:
                    yaml_str = sample.to_yaml(
                        include=sample.get_input_fields() + sample.get_output_fields()
                    )
                    printc(f"yaml_str: {yaml_str}")
                sample_strs.append(yaml_str + "\n")
            except Exception as e:
                print(f"Error: {e} to yaml for {sample}")
                sample_strs.append(str(sample))
        return "\n".join(sample_strs)

    def propose(self):
        r"""Proposing a value while keeping previous value saved on parameter."""
        self._pre_check()
        if self.proposing:
            raise ValueError("Already proposing a value.")
        for demo_param in self.params:
            if demo_param.requires_opt:
                augmented_demos = demo_param._traces
                demos = demo_param._student_traces

                if len(augmented_demos) != len(demos):
                    log.warning(
                        f"augmented and raw demos must have the same length, got {len(augmented_demos)} and {len(demos)} \n {augmented_demos} \n and student demos {demos}"
                    )
                try:
                    sampled_augmented_demos, sampled_raw_demos = self.sample(
                        augmented_demos=augmented_demos,
                        demos=demos,
                        dataset=self.dataset,
                        raw_shots=self._raw_shots,
                        bootstrap_shots=self._bootstrap_shots,
                        weighted=self._weighted,
                    )
                    print(
                        f"sampled_augmented_demos: {[demo.id for demo in sampled_augmented_demos]}"
                    )
                    samples = sampled_augmented_demos + sampled_raw_demos

                    demo_str = ""
                    if len(sampled_augmented_demos) > 0:

                        demo_str = self.samples_to_str(
                            samples=sampled_augmented_demos,
                            augmented=True,
                            exclude_inputs=self.exclude_input_fields_from_bootstrap_demos,
                        )
                    if len(sampled_raw_demos) > 0:
                        demo_str += "\n" + self.samples_to_str(
                            samples=sampled_raw_demos, augmented=False
                        )
                    demo_str = demo_str.strip()
                    demo_param.propose_data(demo_str, samples)
                except Exception as e:
                    print(f"Error: {e} for {demo_param.name}")
                    raise e

        self.proposing = True

    def revert(self):
        """Revert to the previous value when the evaluation is worse."""
        self._pre_check()
        if not self.proposing:
            raise ValueError("Not proposing a value.")
        for param in self.params:
            param.revert_data(include_demos=True)
        self.proposing = False

    def step(self):
        """Discard the previous value and keep the proposed value."""
        self._pre_check()
        if not self.proposing:
            raise ValueError("Not proposing a value.")
        for param in self.params:
            param.step_data(include_demos=True)
            # TODO: track all past history

        self.proposing = False

    def __str__(self) -> str:
        s = f"BootstrapFewShot(raw_shots={self._raw_shots}, bootstrap_shots={self._bootstrap_shots}, \
            params={[p.name for p in self.params]}, dataset={len(self.dataset) if self.dataset else 0})"
        return s

    def __repr__(self) -> str:
        return self.__str__()



"""Reranking model using modelclient as a retriever."""

from typing import List, Optional, Callable, Any, Dict
import logging

from adalflow.core.retriever import (
    Retriever,
)
from adalflow.core.types import (
    RetrieverStrQueriesType,
    RetrieverOutputType,
    RetrieverDocumentsType,
    RetrieverOutput,
    ModelType,
)
from adalflow.core.model_client import ModelClient

log = logging.getLogger(__name__)


class RerankerRetriever(Retriever[str, RetrieverStrQueriesType]):
    __doc__ = r"""
    A retriever that uses a reranker model to rank the documents and retrieve the top-k documents.

    Args:
        top_k (int, optional): The number of top documents to retrieve. Defaults to 5.
        model_client (ModelClient): The model client that has a reranker model,
            such as ``CohereAPIClient`` or ``TransformersClient``.
        model_kwargs (Dict): The model kwargs to pass to the model client.
        documents (Optional[RetrieverDocumentsType], optional): The documents to build the index from. Defaults to None.
        document_map_func (Optional[Callable[[Any], str]], optional): The function to map the document of Any type to the specific type ``RetrieverDocumentType`` that the retriever expects. Defaults to None.

    Examples:


    """

    def __init__(
        self,
        model_client: ModelClient,  # make sure you initialize the model client first
        model_kwargs: Dict = {},
        top_k: int = 5,
        documents: Optional[RetrieverDocumentsType] = None,
        document_map_func: Optional[Callable[[Any], str]] = None,
    ):
        super().__init__()
        self.top_k = top_k
        self._model_kwargs = model_kwargs or {}
        assert "model" in self._model_kwargs, "model must be specified in model_kwargs"
        if not isinstance(self._model_kwargs, Dict):
            raise TypeError(
                f"{type(self).__name__} requires a dictionary for model_kwargs, not a string"
            )
        if not isinstance(model_client, ModelClient):
            raise TypeError(
                f"{type(self).__name__} requires a ModelClient instance for model_client"
            )
        self.model_client = model_client

        self.reset_index()
        if documents:
            self.build_index_from_documents(documents, document_map_func)

    def reset_index(self):
        self.indexed = False
        self.documents = []
        self.total_documents: int = 0

    def build_index_from_documents(
        self,
        documents: RetrieverDocumentsType,
        document_map_func: Optional[Callable[[Any], str]] = None,
    ):
        if document_map_func:
            documents = [document_map_func(doc) for doc in documents]
        else:
            documents = documents
        self.total_documents = len(documents)

        self._model_kwargs["documents"] = documents

        self.indexed = True

    def call(
        self, input: RetrieverStrQueriesType, top_k: Optional[int] = None
    ) -> RetrieverOutputType:
        top_k = top_k or self.top_k
        queries = input if isinstance(input, List) else [input]
        retrieved_outputs: RetrieverOutputType = []

        model_kwargs = self._model_kwargs.copy()
        model_kwargs["top_k"] = top_k

        for query in queries:
            api_kwargs = self.model_client.convert_inputs_to_api_kwargs(
                input=query,
                model_kwargs=model_kwargs,
                model_type=ModelType.RERANKER,
            )
            log.info(f"api_kwargs: {api_kwargs}")
            top_k_indices, top_k_scores = self.model_client.call(
                api_kwargs=api_kwargs, model_type=ModelType.RERANKER
            )
            retrieved_outputs.append(
                RetrieverOutput(
                    doc_indices=top_k_indices,
                    doc_scores=top_k_scores,
                    query=query,
                )
            )
        return retrieved_outputs

    def _extra_repr(self) -> str:
        s = f"top_k={self.top_k}, model_kwargs={self._model_kwargs}, model_client={self.model_client}, total_documents={self.total_documents}"
        return s



"""Define the data model for the SQLAlchemy ORM. This is the schema for the database.

We use sqlalchemy as it is a popular ORM for Python and supports a wide range of databases.

For example:

# Connect to PostgreSQL
engine = create_engine('postgresql://user:password@localhost/mydatabase')

# Connect to MySQL
engine = create_engine('mysql+pymysql://user:password@localhost/mydatabase')

# Connect to SQLite (local file)
engine = create_engine('sqlite:///path/to/database.db')

# Connect to SQLite (in-memory)
engine = create_engine('sqlite:///:memory:')

We define the schema for the database using the declarative_base class provided by sqlalchemy. This allows us to define our data model as a class and then create the database schema from it.

Note:
    - This script might only be applicable to PostgreSQL databases along with the pgvector extension.
    - Ensure you have set up the right database and that it supports the pgvector extension, either local or cloud.

References:
[1] https://github.com/pgvector/pgvector?tab=readme-ov-file#installation-notes---linux-and-mac
[2] pgvector python package: https://pypi.org/project/pgvector/
"""

from typing import Optional, Dict, List
from datetime import datetime
import logging
from adalflow.utils.lazy_import import safe_import, OptionalPackages

sqlalchemy = safe_import(
    OptionalPackages.SQLALCHEMY.value[0], OptionalPackages.SQLALCHEMY.value[1]
)
pgvector = safe_import(
    OptionalPackages.PGVECTOR.value[0], OptionalPackages.PGVECTOR.value[1]
)
from sqlalchemy import (
    Column,
    Integer,
    String,
    Text,
    DateTime,
    TIMESTAMP,
    func,
)
import uuid


from sqlalchemy.orm import Session as BaseSession
from sqlalchemy.dialects.postgresql import JSONB

from pgvector.sqlalchemy import Vector

from adalflow.database.sqlalchemy.base import Base


log = logging.getLogger(__name__)


class DocumentBase(Base):
    # __tablename__ = "document"
    __table_args__ = {"extend_existing": True}  # Allows extending the table
    __abstract__ = True
    id = Column(String, primary_key=True, default=lambda: str(uuid.uuid4()))
    text = Column(Text, nullable=False)

    meta_data = Column(
        JSONB
    )  # Utilizing JSONB for metadata storage if using PostgreSQL
    order = Column(Integer)
    parent_doc_id = Column(String)
    estimated_num_tokens = Column(Integer)
    created_at = Column(
        DateTime, default=datetime.utcnow
    )  # Sets the timestamp on record creation
    updated_at = Column(
        DateTime, default=datetime.utcnow, onupdate=datetime.utcnow
    )  # Updates timestamp on any update

    def to_dict(self):
        return {
            "id": self.id,
            "text": self.text,
            "meta_data": self.meta_data,
            "order": self.order,
            "parent_doc_id": self.parent_doc_id,
            "estimated_num_tokens": self.estimated_num_tokens,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
        }

    @classmethod
    def insert_update_single(cls, document_data: Dict, session: BaseSession):
        r"""Insert or update a single document in the database."""
        identifier = document_data.get("id", None)
        existing_document = None
        if identifier:  # we can check existing documents
            existing_document = session.query(cls).filter_by(id=identifier).first()

        try:
            if existing_document:
                for key, value in document_data.items():
                    setattr(existing_document, key, value)
                log.debug(f"Updated document: {existing_document}")
            else:
                new_document = cls(**document_data)
                session.add(new_document)
                log.debug(f"Inserted document: {new_document}")
            session.commit()
            return existing_document if existing_document else new_document
        except Exception as e:
            log.error(f"Error: {e}, Rolling back transaction.")
            session.rollback()
            return None

    @classmethod
    def insert_update_bulk(cls, document_datas: List[Dict], session: BaseSession):
        r"""Insert or update multiple documents in the database."""
        try:
            identifiers = [data["id"] for data in document_datas if "id" in data]
            existing_documents = {
                doc.id: doc
                for doc in session.query(cls).filter(cls.id.in_(identifiers)).all()
            }

            for document_data in document_datas:
                existing_document = existing_documents.get(document_data["id"])
                if existing_document:
                    for key, value in document_data.items():
                        setattr(existing_document, key, value)
                else:
                    new_document = cls(**document_data)
                    session.add(new_document)

            session.commit()
            print("Bulk insert/update successful.")
            log.debug("Bulk insert/update successful.")
        except Exception as e:
            print(f"Error during bulk insert/update: {e}")
            log.error(f"Error: {e}, Rolling back transaction for bulk insert/update.")
            session.rollback()
            return None


# TODO: extend more fields to support full-text search
# class DocumentSearchExtended(DocumentBase):
#     __abstract__ = True
#     text_tsvector = Column("text_tsvector", String)


class DocumentWithoutVector(DocumentBase):
    __doc__ = r"""Document class without pgvector support."""
    __tablename__ = "document"
    __bind_key__ = "document_db"
    __table_args__ = {"extend_existing": True}  # Allows extending the table

    def __init__(
        self,
        table_name: Optional[str] = None,
        bind_key: Optional[str] = None,
        *args,
        **kwargs,
    ):
        r"""Initialize the DocumentWithoutVector class.

        We allow you to specify the table name and bind key for the document class.
        """
        if table_name:
            type(self).__tablename__ = table_name
        if bind_key:
            type(self).__bind_key__ = bind_key
        super().__init__(*args, **kwargs)


class DocumentModel(DocumentBase):
    __doc__ = r"""Document class extending the DocumentBase with pgvector support."""

    __tablename__ = "document"
    __bind_key__ = "vector_db"

    __table_args__ = {"extend_existing": True}  # Allows extending the table

    vector = Column(Vector())  # using pgvector for vector storage

    def __init__(
        self,
        table_name: Optional[str] = None,
        bind_key: Optional[str] = None,
        *args,
        **kwargs,
    ):
        if table_name:
            type(self).__tablename__ = table_name
        if bind_key:
            type(self).__bind_key__ = bind_key
        super().__init__(*args, **kwargs)


class DialogueTurnModel(Base):
    __tablename__ = "dialogue_turn_all_in_one"
    __bind_key__ = "dialogue_db"

    id = Column(
        String, primary_key=True, default=lambda: str(uuid.uuid4())
    )  # Assuming SERIAL is auto-incrementing integer
    turn_index = Column(Integer)
    user_id = Column(
        String,
        nullable=False,
    )
    session_id = Column(
        String,
        # ForeignKey("dialogue_session.id", ondelete="CASCADE"),
        nullable=False,
    )
    user_query_params = Column(JSONB)
    converse_agent_response = Column(JSONB)
    search_agent_response = Column(JSONB)
    user_feedback = Column(JSONB)
    turn_start_time = Column(DateTime, default=datetime.now())
    converse_agent_response_time = Column(DateTime, default=datetime.now())
    search_agent_response_time = Column(DateTime, default=datetime.now())  # matches
    # matches should be a list of dictionaries
    created_at = Column(TIMESTAMP, default=func.now())
    updated_at = Column(TIMESTAMP, default=func.now(), onupdate=func.now())
    deleted_at = Column(TIMESTAMP, nullable=True)  # soft delete

    def __str__(self):
        str_repr = ""
        str_repr += f"id: {self.id}\n"
        str_repr += f"turn_index: {self.turn_index}\n"
        str_repr += f"user_id: {self.user_id}\n"
        str_repr += f"session_id: {self.session_id}\n"
        str_repr += f"user_query_params: {self.user_query_params}\n"
        str_repr += f"converse_agent_response: {self.converse_agent_response}\n"
        str_repr += f"search_agent_response: {self.search_agent_response}\n"
        str_repr += f"user_feedback: {self.user_feedback}\n"
        str_repr += f"turn_start_time: {self.turn_start_time}\n"
        str_repr += (
            f"converse_agent_response_time: {self.converse_agent_response_time}\n"
        )
        str_repr += f"search_agent_response_time: {self.search_agent_response_time}\n"
        return str_repr

    def to_dict(self):
        return {
            "turn_index": self.turn_index,
            "user_id": self.user_id,
            "session_id": self.session_id,
            "user_query_params": self.user_query_params,
            "converse_agent_response": self.converse_agent_response,
            "search_agent_response": self.search_agent_response,
            "user_feedback": self.user_feedback,
            "turn_start_time": self.turn_start_time,
            "converse_agent_response_time": self.converse_agent_response_time,
            "search_agent_response_time": self.search_agent_response_time,
            "status": self.status,
            "created_at": self.created_at,
            "updated_at": self.updated_at,
            "deleted_at": self.deleted_at,
        }


# if __name__ == "__main__":
#     from sqlalchemy.sql import text

#     from adalflow.database.sqlalchemy.sqlachemy_manager import DatabaseManager

#     db_name = "vector_db"
#     postgres_url = f"postgresql://postgres:password@localhost:5432/{db_name}"

#     db_manager = DatabaseManager(postgres_url)
#     engine = db_manager.engine
#     table_names = db_manager.list_table_schemas()
#     print(f"Table names: {table_names}")
#     database_schema = db_manager.list_database_schemas()
#     print(f"Database schema: {database_schema}")
#     database_name = db_manager.get_database_name()
#     print(f"Database name: {database_name}")
#     # engine = create_engine(postgres_url)

#     try:
#         # Begin a transaction
#         with engine.connect() as connection:
#             # Use the connection to execute a raw SQL command
#             # Ensuring the session commits after, especially for operations requiring administrative rights
#             connection.execute(text("CREATE EXTENSION IF NOT EXISTS vector;"))
#             connection.commit()  # Explicitly commit to ensure that the CREATE EXTENSION command is finalized
#     except Exception as e:
#         print(f"An error occurred while trying to create the extension: {e}")

#     # Base.metadata.create_all(engine, checkfirst=True)
#     # Session = sessionmaker(bind=engine)
#     # session = Session()
#     db_manager.create_tables()
#     session = db_manager.get_session()

#     # doc = Document(text="Hello world")
#     DocumentModel.insert_update_single({"text": "Hello world"}, session=session)
#     # do bulk insert
#     DocumentModel.insert_update_bulk(
#         [
#             {"id": "doc2", "text": "Hello world 2"},
#             {"id": "doc3", "text": "Hello world 3"},
#         ],
#         session=session,
#     )
#     db_manager.close()
#     # session.add(doc)
#     # session.commit()
#     # from adalflow.core.types import Document as LightragDocument

#     # doc_2 = LightragDocument.from_dict(doc.to_dict())
#     # print("doc:", doc, doc.id)
#     # print("doc_2:", doc_2, doc_2.id)

#     # print(doc.id)
#     # print(doc.text)
#     # session.close()
#     # engine.dispose()



"""Leverage a postgres database to store and retrieve documents."""

from typing import List, Optional, Any
from enum import Enum
import numpy as np
import logging

from adalflow.core.retriever import (
    Retriever,
)
from adalflow.core.embedder import Embedder

from adalflow.core.types import (
    RetrieverOutput,
    RetrieverStrQueryType,
    RetrieverStrQueriesType,
    Document,
)
from adalflow.database.sqlalchemy.sqlachemy_manager import DatabaseManager

log = logging.getLogger(__name__)


class DistanceToOperator(Enum):
    __doc__ = r"""Enum for the distance to operator.

    About pgvector:

    1. L2 distance: <->, inner product (<#>), cosine distance (<=>), and L1 distance (<+>, added in 0.7.0)
    """
    L2 = "<->"
    INNER_PRODUCT = (
        "<#>"  # cosine similarity when the vector is normalized, in range [-1, 1]
    )
    COSINE = "<=>"  # cosine distance, in range [0, 1] = 1 - cosine_similarity
    L1 = "<+>"


class PostgresRetriever(Retriever[Any, RetrieverStrQueryType]):
    __doc__ = r"""Use a postgres database to store and retrieve documents.

    Users can follow this example and to customize the prompt or additionally ask it to output score along with the indices.

    Args:
        top_k (Optional[int], optional): top k documents to fetch. Defaults to 1.
        database_url (str): the database url to connect to. Defaults to postgresql://postgres:password@localhost:5432/vector_db.

    References:
    [1] pgvector extension: https://github.com/pgvector/pgvector
    """

    def __init__(
        self,
        embedder: Embedder,
        top_k: Optional[int] = 1,
        database_url: str = None,
        table_name: str = "document",
        distance_operator: DistanceToOperator = DistanceToOperator.INNER_PRODUCT,
    ):
        super().__init__()
        self.top_k = top_k
        self.table_name = table_name
        db_name = "vector_db"
        self.database_url = (
            database_url or f"postgresql://postgres:password@localhost:5432/{db_name}"
        )
        self.db_manager = DatabaseManager(self.database_url)
        self.embedder = embedder
        self.distance_operator = distance_operator
        self.db_score_prob_fun_map = {
            DistanceToOperator.COSINE: self._convert_cosine_distance_to_probability,
            DistanceToOperator.L2: self._convert_l2_distance_to_probability,
            DistanceToOperator.INNER_PRODUCT: self._convert_cosine_similarity_to_probability,
        }
        self.score_prob_fun = (
            self.db_score_prob_fun_map[self.distance_operator]
            if self.distance_operator in self.db_score_prob_fun_map
            else None
        )

    @classmethod
    def format_vector_search_query(
        cls,
        table_name: str,
        vector_column: str,
        query_embedding: List[float],
        top_k: int,
        distance_operator: DistanceToOperator,
        sort_desc: bool = True,
    ) -> str:
        """
        Formats a SQL query string to select all columns from a table, order the results
        by the distance or similarity score to a provided embedding, and also return
        that score.

        Args:
            table_name (str): The name of the table to query.
            column (str): The name of the column containing the vector data.
            query_embedding (list or str): The embedding vector to compare against.
            top_k (int): The number of top results to return.

        Returns:
            str: A formatted SQL query string that includes the score.
        """

        # Convert the list embedding to a string format suitable for SQL
        if isinstance(query_embedding, list):
            embedding_str = str(query_embedding).replace(
                " ", ""
            )  # Remove spaces for cleaner SQL
        else:
            embedding_str = query_embedding

        # Determine sorting order
        order_by = "DESC" if sort_desc else "ASC"

        # SQL query that includes the score in the selected columns
        sql_query = f"""
        SELECT *, ({vector_column} {distance_operator.value} '{embedding_str}') AS score
        FROM {table_name}
        ORDER BY score {order_by}
        LIMIT {top_k};
        """
        return sql_query

    def retrieve_by_sql(self, query: str) -> List[str]:
        """Retrieve documents from the postgres database."""

        results = self.db_manager.execute_query(query)
        print(results)
        return results

    def _convert_cosine_similarity_to_probability(
        self, cosine_similarity: List[float]
    ) -> List[float]:
        """Convert cosine similarity to probability."""
        return [(1 + cosine_similarity) / 2 for cosine_similarity in cosine_similarity]

    def _convert_l2_distance_to_probability(
        self, l2_distance: List[float]
    ) -> List[float]:
        """Convert L2 distance to probability.

        note:

        Ensure the vector is normalized so that the l2_distance will be in range [0, 2]
        """
        distance = np.array(l2_distance)
        # clip to ensure the distance is in range [0, 2]
        distance = np.clip(distance, 0, 2)
        # convert to probability
        prob_score = 1 - distance / 2
        return prob_score.tolist()

    def _convert_cosine_distance_to_probability(
        self, cosine_distance: List[float]
    ) -> List[float]:
        """Convert cosine distance to probability."""
        return [(1 - cosine_distance) for cosine_distance in cosine_distance]

    def call(
        self,
        input: RetrieverStrQueriesType,
        top_k: Optional[int] = None,
        **kwargs,
    ) -> List[RetrieverOutput]:
        top_k = top_k or self.top_k
        queries: List[str] = input if isinstance(input, list) else [input]
        retrieved_outputs: List[RetrieverOutput] = []
        queries_embeddings = self.embedder(queries)

        sort_desc = False
        if (
            self.distance_operator == DistanceToOperator.INNER_PRODUCT
        ):  # cosine similarity
            sort_desc = True

        for idx, query in enumerate(queries):
            query_embedding = queries_embeddings.data[idx].embedding
            query_str = self.format_vector_search_query(
                table_name=self.table_name,
                vector_column="vector",
                query_embedding=query_embedding,
                top_k=top_k,
                distance_operator=self.distance_operator,
                sort_desc=sort_desc,
            )
            retrieved_documents = self.retrieve_by_sql(query_str)
            doc_indices = [doc["id"] for doc in retrieved_documents]
            doc_scores = [doc["score"] for doc in retrieved_documents]
            doc_scores_prob = (
                self.score_prob_fun(doc_scores) if self.score_prob_fun else None
            )
            documents: List[Document] = []
            for doc in retrieved_documents:
                documents.append(Document.from_dict(doc))
            retrieved_outputs.append(
                RetrieverOutput(
                    doc_indices=doc_indices,
                    doc_scores=doc_scores_prob if doc_scores_prob else doc_scores,
                    query=query,
                    documents=documents,
                )
            )
        return retrieved_outputs


# if __name__ == "__main__":
#     from adalflow.core.embedder import Embedder
#     from adalflow.core.types import Document
#     from adalflow.database.sqlalchemy.pipeline.default_config import default_config
#     from adalflow.utils import setup_env  # noqa

#     documents = [
#         {
#             "meta_data": {"title": "Li Yin's profile"},
#             "text": "My name is Li Yin, I love rock climbing"
#             + "lots of nonsense text" * 500,
#             "id": "doc1",
#         },
#         {
#             "meta_data": {"title": "Interviewing Li Yin"},
#             "text": "lots of more nonsense text" * 250
#             + "Li Yin is a software developer and AI researcher"
#             + "lots of more nonsense text" * 250,
#             "id": "doc2",
#         },
#     ]
#     db_name = "vector_db"
#     postgres_url = f"postgresql://postgres:password@localhost:5432/{db_name}"

#     vector_config = default_config["to_embeddings"]["component_config"]["embedder"][
#         "component_config"
#     ]
#     eb = Embedder.from_config(vector_config)
#     pr = PostgresRetriever(
#         embedder=eb,
#         database_url=postgres_url,
#         top_k=2,
#         distance_operator=DistanceToOperator.INNER_PRODUCT,
#     )
#     output = pr("What did the author do?")
#     print(output)



r"""The component that orchestrates model client (Embedding models in particular) and output processors."""

from typing import Optional, Any, Dict, List
import logging
from tqdm import tqdm

from adalflow.core.types import ModelType, EmbedderOutput
from adalflow.core.model_client import ModelClient
from adalflow.core.types import (
    EmbedderOutputType,
    EmbedderInputType,
    BatchEmbedderInputType,
    BatchEmbedderOutputType,
)
from adalflow.core.component import Component
import adalflow.core.functional as F

__all__ = ["Embedder", "BatchEmbedder"]

log = logging.getLogger(__name__)


class Embedder(Component):
    r"""
    A user-facing component that orchestrates an embedder model via the model client and output processors.

    Args:
        model_client (ModelClient): The model client to use for the embedder.
        model_kwargs (Dict[str, Any], optional): The model kwargs to pass to the model client. Defaults to {}.
        output_processors (Optional[Component], optional): The output processors after model call. Defaults to None.
            If you want to add further processing, it should operate on the ``EmbedderOutput`` data type.

    input: a single str or a list of str. When a list is used, the list is processed as a batch of inputs in the model client.

    Note:
        - The ``output_processors`` will be applied only on the data field of ``EmbedderOutput``, which is a list of ``Embedding``.
        - Use ``BatchEmbedder`` for automatically batching input of large size, larger than 100.
    """

    model_type: ModelType = ModelType.EMBEDDER
    model_client: ModelClient
    output_processors: Optional[Component]

    def __init__(
        self,
        *,
        model_client: ModelClient,
        model_kwargs: Dict[str, Any] = {},
        output_processors: Optional[Component] = None,
    ) -> None:

        super().__init__(model_kwargs=model_kwargs)
        if not isinstance(model_kwargs, Dict):
            raise TypeError(
                f"{type(self).__name__} requires a dictionary for model_kwargs, not a string"
            )
        self.model_kwargs = model_kwargs.copy()

        if not isinstance(model_client, ModelClient):
            raise TypeError(
                f"{type(self).__name__} requires a ModelClient instance for model_client, please pass it as OpenAIClient() or GroqAPIClient() for example."
            )
        self.model_client = model_client
        self.output_processors = output_processors

    @classmethod
    def from_config(cls, config: Dict[str, Any]) -> "Embedder":
        """Create an Embedder from a configuration dictionary.

        Example:

        .. code-block:: python

            embedder_config =  {
                "model_client": {
                    "component_name": "OpenAIClient",
                    "component_config": {}
                },
                "model_kwargs": {
                    "model": "text-embedding-3-small",
                    "dimensions": 256,
                    "encoding_format": "float"
                }
            }

            embedder = Embedder.from_config(embedder_config)
        """
        if "model_client" not in config:
            raise ValueError("model_client is required in the config")
        return super().from_config(config)

    def _compose_model_kwargs(self, **model_kwargs) -> Dict[str, object]:
        r"""Add new arguments or overwrite existing arguments in the model_kwargs."""
        return F.compose_model_kwargs(self.model_kwargs, model_kwargs)

    def _pre_call(
        self, input: EmbedderInputType, model_kwargs: Optional[Dict] = {}
    ) -> Dict:
        # step 1: combine the model_kwargs with the default model_kwargs
        composed_model_kwargs = self._compose_model_kwargs(**model_kwargs)
        # step 2: convert the input to the api_kwargs
        api_kwargs = self.model_client.convert_inputs_to_api_kwargs(
            input=input,
            model_kwargs=composed_model_kwargs,
            model_type=self.model_type,
        )
        log.debug(f"api_kwargs: {api_kwargs}")
        return api_kwargs

    def _post_call(self, response: Any) -> EmbedderOutputType:
        r"""Get float list response and process it with output_processor"""
        try:
            embedding_output: EmbedderOutputType = (
                self.model_client.parse_embedding_response(response)
            )
        except Exception as e:
            log.error(f"Error parsing the embedding {response}: {e}")
            return EmbedderOutput(raw_response=str(response), error=str(e))
        output: EmbedderOutputType = EmbedderOutputType(raw_response=embedding_output)
        # data = embedding_output.data
        if self.output_processors:
            try:
                embedding_output = self.output_processors(embedding_output)
                output.data = embedding_output
            except Exception as e:
                log.error(f"Error processing the output: {e}")
                output.error = str(e)
        else:
            output.data = embedding_output.data

        return output

    def call(
        self,
        input: EmbedderInputType,
        model_kwargs: Optional[Dict] = {},
    ) -> EmbedderOutputType:
        log.debug(f"Calling {self.__class__.__name__} with input: {input}")
        api_kwargs = self._pre_call(input=input, model_kwargs=model_kwargs)
        output: EmbedderOutputType = None
        response = None
        try:
            response = self.model_client.call(
                api_kwargs=api_kwargs, model_type=self.model_type
            )
        except Exception as e:
            log.error(f"Error calling the model: {e}")
            output = EmbedderOutput(error=str(e))

        if response:
            try:
                output = self._post_call(response)
            except Exception as e:
                log.error(f"Error processing output: {e}")
                output = EmbedderOutput(raw_response=str(response), error=str(e))

        # add back the input
        output.input = [input] if isinstance(input, str) else input
        log.debug(f"Output from {self.__class__.__name__}: {output}")
        return output

    async def acall(
        self,
        input: EmbedderInputType,
        model_kwargs: Optional[Dict] = {},
    ) -> EmbedderOutputType:
        log.debug(f"Calling {self.__class__.__name__} with input: {input}")
        api_kwargs = self._pre_call(input=input, model_kwargs=model_kwargs)
        output: EmbedderOutputType = None
        response = None
        try:
            response = await self.model_client.acall(
                api_kwargs=api_kwargs, model_type=self.model_type
            )
        except Exception as e:
            log.error(f"Error calling the model: {e}")
            output = EmbedderOutput(error=str(e))

        if response:
            try:
                output = self._post_call(response)
            except Exception as e:
                log.error(f"Error processing output: {e}")
                output = EmbedderOutput(raw_response=str(response), error=str(e))
        # add back the input
        output.input = [input] if isinstance(input, str) else input
        log.debug(f"Output from {self.__class__.__name__}: {output}")
        return output

    def _extra_repr(self) -> str:
        s = f"model_kwargs={self.model_kwargs}, "
        return s


class BatchEmbedder(Component):
    __doc__ = r"""Adds batching to the embedder component.

    Args:
        embedder (Embedder): The embedder to use for batching.
        batch_size (int, optional): The batch size to use for batching. Defaults to 100.
    """

    def __init__(self, embedder: Embedder, batch_size: int = 100) -> None:
        super().__init__(batch_size=batch_size)
        self.embedder = embedder
        self.batch_size = batch_size

    def call(
        self, input: BatchEmbedderInputType, model_kwargs: Optional[Dict] = {}
    ) -> BatchEmbedderOutputType:
        r"""Call the embedder with batching.

        Args:
            input (BatchEmbedderInputType): The input to the embedder. Use this when you have a large input that needs to be batched. Also ensure
            the output can fit into memory.
            model_kwargs (Optional[Dict], optional): The model kwargs to pass to the embedder. Defaults to {}.

        Returns:
            BatchEmbedderOutputType: The output from the embedder.
        """

        if isinstance(input, str):
            input = [input]
        n = len(input)
        embeddings: List[EmbedderOutputType] = []
        for i in tqdm(
            range(0, n, self.batch_size),
            desc="Batch embedding documents",
        ):
            batch_input = input[i : i + self.batch_size]
            batch_output = self.embedder.call(
                input=batch_input, model_kwargs=model_kwargs
            )
            embeddings.append(batch_output)
        return embeddings



"""OpenAI ModelClient integration."""

import os
import base64
from typing import (
    Dict,
    Sequence,
    Optional,
    List,
    Any,
    TypeVar,
    Callable,
    Generator,
    Union,
    Literal,
)
import re

import logging
import backoff

# optional import
from adalflow.utils.lazy_import import safe_import, OptionalPackages


openai = safe_import(OptionalPackages.OPENAI.value[0], OptionalPackages.OPENAI.value[1])

from openai import OpenAI, AsyncOpenAI, Stream
from openai import (
    APITimeoutError,
    InternalServerError,
    RateLimitError,
    UnprocessableEntityError,
    BadRequestError,
)
from openai.types import (
    Completion,
    CreateEmbeddingResponse,
    Image,
)
from openai.types.chat import ChatCompletionChunk, ChatCompletion

from adalflow.core.model_client import ModelClient
from adalflow.core.types import (
    ModelType,
    EmbedderOutput,
    TokenLogProb,
    CompletionUsage,
    GeneratorOutput,
)
from adalflow.components.model_client.utils import parse_embedding_response

log = logging.getLogger(__name__)
T = TypeVar("T")


# completion parsing functions and you can combine them into one singple chat completion parser
def get_first_message_content(completion: ChatCompletion) -> str:
    r"""When we only need the content of the first message.
    It is the default parser for chat completion."""
    return completion.choices[0].message.content


# def _get_chat_completion_usage(completion: ChatCompletion) -> OpenAICompletionUsage:
#     return completion.usage


def parse_stream_response(completion: ChatCompletionChunk) -> str:
    r"""Parse the response of the stream API."""
    return completion.choices[0].delta.content


def handle_streaming_response(generator: Stream[ChatCompletionChunk]):
    r"""Handle the streaming response."""
    for completion in generator:
        log.debug(f"Raw chunk completion: {completion}")
        parsed_content = parse_stream_response(completion)
        yield parsed_content


def get_all_messages_content(completion: ChatCompletion) -> List[str]:
    r"""When the n > 1, get all the messages content."""
    return [c.message.content for c in completion.choices]


def get_probabilities(completion: ChatCompletion) -> List[List[TokenLogProb]]:
    r"""Get the probabilities of each token in the completion."""
    log_probs = []
    for c in completion.choices:
        content = c.logprobs.content
        print(content)
        log_probs_for_choice = []
        for openai_token_logprob in content:
            token = openai_token_logprob.token
            logprob = openai_token_logprob.logprob
            log_probs_for_choice.append(TokenLogProb(token=token, logprob=logprob))
        log_probs.append(log_probs_for_choice)
    return log_probs


class OpenAIClient(ModelClient):
    __doc__ = r"""A component wrapper for the OpenAI API client.

    Support both embedding and chat completion API, including multimodal capabilities.

    Users (1) simplify use ``Embedder`` and ``Generator`` components by passing OpenAIClient() as the model_client.
    (2) can use this as an example to create their own API client or extend this class(copying and modifing the code) in their own project.

    Note:
        We suggest users not to use `response_format` to enforce output data type or `tools` and `tool_choice`  in your model_kwargs when calling the API.
        We do not know how OpenAI is doing the formating or what prompt they have added.
        Instead
        - use :ref:`OutputParser<components-output_parsers>` for response parsing and formating.

        For multimodal inputs, provide images in model_kwargs["images"] as a path, URL, or list of them.
        The model must support vision capabilities (e.g., gpt-4o, gpt-4o-mini, o1, o1-mini).

        For image generation, use model_type=ModelType.IMAGE_GENERATION and provide:
        - model: "dall-e-3" or "dall-e-2"
        - prompt: Text description of the image to generate
        - size: "1024x1024", "1024x1792", or "1792x1024" for DALL-E 3; "256x256", "512x512", or "1024x1024" for DALL-E 2
        - quality: "standard" or "hd" (DALL-E 3 only)
        - n: Number of images to generate (1 for DALL-E 3, 1-10 for DALL-E 2)
        - response_format: "url" or "b64_json"

    Args:
        api_key (Optional[str], optional): OpenAI API key. Defaults to None.
        chat_completion_parser (Callable[[Completion], Any], optional): A function to parse the chat completion to a str. Defaults to None.
            Default is `get_first_message_content`.

    References:
        - Embeddings models: https://platform.openai.com/docs/guides/embeddings
        - Chat models: https://platform.openai.com/docs/guides/text-generation
        - Vision models: https://platform.openai.com/docs/guides/vision
        - Image models: https://platform.openai.com/docs/guides/images
        - OpenAI docs: https://platform.openai.com/docs/introduction
    """

    def __init__(
        self,
        api_key: Optional[str] = None,
        chat_completion_parser: Callable[[Completion], Any] = None,
        input_type: Literal["text", "messages"] = "text",
    ):
        r"""It is recommended to set the OPENAI_API_KEY environment variable instead of passing it as an argument.

        Args:
            api_key (Optional[str], optional): OpenAI API key. Defaults to None.
        """
        super().__init__()
        self._api_key = api_key
        self.sync_client = self.init_sync_client()
        self.async_client = None  # only initialize if the async call is called
        self.chat_completion_parser = (
            chat_completion_parser or get_first_message_content
        )
        self._input_type = input_type

    def init_sync_client(self):
        api_key = self._api_key or os.getenv("OPENAI_API_KEY")
        if not api_key:
            raise ValueError("Environment variable OPENAI_API_KEY must be set")
        return OpenAI(api_key=api_key)

    def init_async_client(self):
        api_key = self._api_key or os.getenv("OPENAI_API_KEY")
        if not api_key:
            raise ValueError("Environment variable OPENAI_API_KEY must be set")
        return AsyncOpenAI(api_key=api_key)

    # def _parse_chat_completion(self, completion: ChatCompletion) -> "GeneratorOutput":
    #     # TODO: raw output it is better to save the whole completion as a source of truth instead of just the message
    #     try:
    #         data = self.chat_completion_parser(completion)
    #         usage = self.track_completion_usage(completion)
    #         return GeneratorOutput(
    #             data=data, error=None, raw_response=str(data), usage=usage
    #         )
    #     except Exception as e:
    #         log.error(f"Error parsing the completion: {e}")
    #         return GeneratorOutput(data=None, error=str(e), raw_response=completion)

    def parse_chat_completion(
        self,
        completion: Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]],
    ) -> "GeneratorOutput":
        """Parse the completion, and put it into the raw_response."""
        log.debug(f"completion: {completion}, parser: {self.chat_completion_parser}")
        try:
            data = self.chat_completion_parser(completion)
            usage = self.track_completion_usage(completion)
            return GeneratorOutput(
                data=None, error=None, raw_response=data, usage=usage
            )
        except Exception as e:
            log.error(f"Error parsing the completion: {e}")
            return GeneratorOutput(data=None, error=str(e), raw_response=completion)

    def track_completion_usage(
        self,
        completion: Union[ChatCompletion, Generator[ChatCompletionChunk, None, None]],
    ) -> CompletionUsage:
        if isinstance(completion, ChatCompletion):
            usage: CompletionUsage = CompletionUsage(
                completion_tokens=completion.usage.completion_tokens,
                prompt_tokens=completion.usage.prompt_tokens,
                total_tokens=completion.usage.total_tokens,
            )
            return usage
        else:
            raise NotImplementedError(
                "streaming completion usage tracking is not implemented"
            )

    def parse_embedding_response(
        self, response: CreateEmbeddingResponse
    ) -> EmbedderOutput:
        r"""Parse the embedding response to a structure Adalflow components can understand.

        Should be called in ``Embedder``.
        """
        try:
            return parse_embedding_response(response)
        except Exception as e:
            log.error(f"Error parsing the embedding response: {e}")
            return EmbedderOutput(data=[], error=str(e), raw_response=response)

    def convert_inputs_to_api_kwargs(
        self,
        input: Optional[Any] = None,
        model_kwargs: Dict = {},
        model_type: ModelType = ModelType.UNDEFINED,
    ) -> Dict:
        r"""
        Specify the API input type and output api_kwargs that will be used in _call and _acall methods.
        Convert the Component's standard input, and system_input(chat model) and model_kwargs into API-specific format.
        For multimodal inputs, images can be provided in model_kwargs["images"] as a string path, URL, or list of them.
        The model specified in model_kwargs["model"] must support multimodal capabilities when using images.

        Args:
            input: The input text or messages to process
            model_kwargs: Additional parameters including:
                - images: Optional image source(s) as path, URL, or list of them
                - detail: Image detail level ('auto', 'low', or 'high'), defaults to 'auto'
                - model: The model to use (must support multimodal inputs if images are provided)
            model_type: The type of model (EMBEDDER or LLM)

        Returns:
            Dict: API-specific kwargs for the model call
        """

        final_model_kwargs = model_kwargs.copy()
        if model_type == ModelType.EMBEDDER:
            if isinstance(input, str):
                input = [input]
            # convert input to input
            if not isinstance(input, Sequence):
                raise TypeError("input must be a sequence of text")
            final_model_kwargs["input"] = input
        elif model_type == ModelType.LLM:
            # convert input to messages
            messages: List[Dict[str, str]] = []
            images = final_model_kwargs.pop("images", None)
            detail = final_model_kwargs.pop("detail", "auto")

            if self._input_type == "messages":
                system_start_tag = "<START_OF_SYSTEM_PROMPT>"
                system_end_tag = "<END_OF_SYSTEM_PROMPT>"
                user_start_tag = "<START_OF_USER_PROMPT>"
                user_end_tag = "<END_OF_USER_PROMPT>"
                pattern = f"{system_start_tag}(.*?){system_end_tag}{user_start_tag}(.*?){user_end_tag}"
                # Compile the regular expression
                regex = re.compile(pattern)
                # Match the pattern
                match = regex.search(input)
                system_prompt, input_str = None, None

                if match:
                    system_prompt = match.group(1)
                    input_str = match.group(2)
                else:
                    print("No match found.")
                if system_prompt and input_str:
                    messages.append({"role": "system", "content": system_prompt})
                    if images:
                        content = [{"type": "text", "text": input_str}]
                        if isinstance(images, (str, dict)):
                            images = [images]
                        for img in images:
                            content.append(self._prepare_image_content(img, detail))
                        messages.append({"role": "user", "content": content})
                    else:
                        messages.append({"role": "user", "content": input_str})
            if len(messages) == 0:
                if images:
                    content = [{"type": "text", "text": input}]
                    if isinstance(images, (str, dict)):
                        images = [images]
                    for img in images:
                        content.append(self._prepare_image_content(img, detail))
                    messages.append({"role": "user", "content": content})
                else:
                    messages.append({"role": "system", "content": input})
            final_model_kwargs["messages"] = messages
        elif model_type == ModelType.IMAGE_GENERATION:
            # For image generation, input is the prompt
            final_model_kwargs["prompt"] = input
            # Ensure model is specified
            if "model" not in final_model_kwargs:
                raise ValueError("model must be specified for image generation")
            # Set defaults for DALL-E 3 if not specified
            final_model_kwargs["size"] = final_model_kwargs.get("size", "1024x1024")
            final_model_kwargs["quality"] = final_model_kwargs.get(
                "quality", "standard"
            )
            final_model_kwargs["n"] = final_model_kwargs.get("n", 1)
            final_model_kwargs["response_format"] = final_model_kwargs.get(
                "response_format", "url"
            )

            # Handle image edits and variations
            image = final_model_kwargs.get("image")
            if isinstance(image, str) and os.path.isfile(image):
                final_model_kwargs["image"] = self._encode_image(image)

            mask = final_model_kwargs.get("mask")
            if isinstance(mask, str) and os.path.isfile(mask):
                final_model_kwargs["mask"] = self._encode_image(mask)
        else:
            raise ValueError(f"model_type {model_type} is not supported")
        return final_model_kwargs

    def parse_image_generation_response(self, response: List[Image]) -> GeneratorOutput:
        """Parse the image generation response into a GeneratorOutput."""
        try:
            # Extract URLs or base64 data from the response
            data = [img.url or img.b64_json for img in response]
            # For single image responses, unwrap from list
            if len(data) == 1:
                data = data[0]
            return GeneratorOutput(
                data=data,
                raw_response=str(response),
            )
        except Exception as e:
            log.error(f"Error parsing image generation response: {e}")
            return GeneratorOutput(data=None, error=str(e), raw_response=str(response))

    @backoff.on_exception(
        backoff.expo,
        (
            APITimeoutError,
            InternalServerError,
            RateLimitError,
            UnprocessableEntityError,
            BadRequestError,
        ),
        max_time=5,
    )
    def call(self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED):
        """
        kwargs is the combined input and model_kwargs.  Support streaming call.
        """
        log.info(f"api_kwargs: {api_kwargs}")
        if model_type == ModelType.EMBEDDER:
            return self.sync_client.embeddings.create(**api_kwargs)
        elif model_type == ModelType.LLM:
            if "stream" in api_kwargs and api_kwargs.get("stream", False):
                log.debug("streaming call")
                self.chat_completion_parser = handle_streaming_response
                return self.sync_client.chat.completions.create(**api_kwargs)
            return self.sync_client.chat.completions.create(**api_kwargs)
        elif model_type == ModelType.IMAGE_GENERATION:
            # Determine which image API to call based on the presence of image/mask
            if "image" in api_kwargs:
                if "mask" in api_kwargs:
                    # Image edit
                    response = self.sync_client.images.edit(**api_kwargs)
                else:
                    # Image variation
                    response = self.sync_client.images.create_variation(**api_kwargs)
            else:
                # Image generation
                response = self.sync_client.images.generate(**api_kwargs)
            return response.data
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @backoff.on_exception(
        backoff.expo,
        (
            APITimeoutError,
            InternalServerError,
            RateLimitError,
            UnprocessableEntityError,
            BadRequestError,
        ),
        max_time=5,
    )
    async def acall(
        self, api_kwargs: Dict = {}, model_type: ModelType = ModelType.UNDEFINED
    ):
        """
        kwargs is the combined input and model_kwargs
        """
        if self.async_client is None:
            self.async_client = self.init_async_client()
        if model_type == ModelType.EMBEDDER:
            return await self.async_client.embeddings.create(**api_kwargs)
        elif model_type == ModelType.LLM:
            return await self.async_client.chat.completions.create(**api_kwargs)
        elif model_type == ModelType.IMAGE_GENERATION:
            # Determine which image API to call based on the presence of image/mask
            if "image" in api_kwargs:
                if "mask" in api_kwargs:
                    # Image edit
                    response = await self.async_client.images.edit(**api_kwargs)
                else:
                    # Image variation
                    response = await self.async_client.images.create_variation(
                        **api_kwargs
                    )
            else:
                # Image generation
                response = await self.async_client.images.generate(**api_kwargs)
            return response.data
        else:
            raise ValueError(f"model_type {model_type} is not supported")

    @classmethod
    def from_dict(cls: type[T], data: Dict[str, Any]) -> T:
        obj = super().from_dict(data)
        # recreate the existing clients
        obj.sync_client = obj.init_sync_client()
        obj.async_client = obj.init_async_client()
        return obj

    def to_dict(self) -> Dict[str, Any]:
        r"""Convert the component to a dictionary."""
        # TODO: not exclude but save yes or no for recreating the clients
        exclude = [
            "sync_client",
            "async_client",
        ]  # unserializable object
        output = super().to_dict(exclude=exclude)
        return output

    def _encode_image(self, image_path: str) -> str:
        """Encode image to base64 string.

        Args:
            image_path: Path to image file.

        Returns:
            Base64 encoded image string.

        Raises:
            ValueError: If the file cannot be read or doesn't exist.
        """
        try:
            with open(image_path, "rb") as image_file:
                return base64.b64encode(image_file.read()).decode("utf-8")
        except FileNotFoundError:
            raise ValueError(f"Image file not found: {image_path}")
        except PermissionError:
            raise ValueError(f"Permission denied when reading image file: {image_path}")
        except Exception as e:
            raise ValueError(f"Error encoding image {image_path}: {str(e)}")

    def _prepare_image_content(
        self, image_source: Union[str, Dict[str, Any]], detail: str = "auto"
    ) -> Dict[str, Any]:
        """Prepare image content for API request.

        Args:
            image_source: Either a path to local image or a URL.
            detail: Image detail level ('auto', 'low', or 'high').

        Returns:
            Formatted image content for API request.
        """
        if isinstance(image_source, str):
            if image_source.startswith(("http://", "https://")):
                return {
                    "type": "image_url",
                    "image_url": {"url": image_source, "detail": detail},
                }
            else:
                base64_image = self._encode_image(image_source)
                return {
                    "type": "image_url",
                    "image_url": {
                        "url": f"data:image/jpeg;base64,{base64_image}",
                        "detail": detail,
                    },
                }
        return image_source


# Example usage:
# if __name__ == "__main__":
#     from adalflow.core import Generator
#     from adalflow.utils import setup_env, get_logger
#
#     log = get_logger(level="DEBUG")
#
#     setup_env()
#     prompt_kwargs = {"input_str": "What is the meaning of life?"}
#
#     gen = Generator(
#         model_client=OpenAIClient(),
#         model_kwargs={"model": "gpt-3.5-turbo", "stream": True},
#     )
#     gen_response = gen(prompt_kwargs)
#     print(f"gen_response: {gen_response}")
#
#     for genout in gen_response.data:
#         print(f"genout: {genout}")



"""Abstract base class for evaluation metrics."""

from typing import Optional, List, Any

from dataclasses import dataclass


@dataclass
class EvaluationResult:
    """Evaluation result."""

    avg_score: float
    per_item_scores: Optional[List[float]] = None
    additional_info: Optional[dict] = None


class BaseEvaluator:
    def __init__(self, *args, **kwargs):
        pass

    def compute_single_item(self, *args, **kwargs) -> float:
        """Compute the score for a single item."""
        raise NotImplementedError("Subclasses must implement this method.")

    # TODO: support multi-threading or async to speed up evaluation
    def compute(self, *args, **kwargs) -> Any:
        """Evaluate a list of predictions and ground truth values. and return overall score and per-item scores."""
        raise NotImplementedError("Subclasses must implement this method.")

    def __call__(self, *args, **kwargs):
        return self.compute(*args, **kwargs)



"""LLM as retriever module."""

from typing import Optional, Any, Dict, Callable
import logging

from adalflow.core.retriever import (
    Retriever,
)
from adalflow.core.generator import Generator
from adalflow.core.model_client import ModelClient
from adalflow.core.string_parser import ListParser
from adalflow.core.types import (
    GeneratorOutput,
    RetrieverOutput,
    RetrieverDocumentsType,
    RetrieverStrQueryType,
    RetrieverStrQueriesType,
    RetrieverOutputType,
)

log = logging.getLogger(__name__)


DEFAULT_LLM_AS_RETRIEVER_PROMPT_TEMPLATE = r"""<SYS>
You are a retriever. Given a list of documents, you will retrieve the top_k {{top_k}} most relevant documents and output the indices (int) as a list:
[<index of the most relevant top_k options>]
<Documents>
{% for doc in documents %}
```Index {{ loop.index - 1 }}. {{ doc }}```
______________
{% endfor %}
</Documents>
</SYS>
Query: {{ input_str }}
You:
"""


class LLMRetriever(Retriever[str, RetrieverStrQueryType]):
    __doc__ = r"""Use LLM to access the query and the documents to retrieve the top k relevant indices of the documents.

    Users can follow this example and to customize the prompt or additionally ask it to output score along with the indices.

    Args:
        top_k (Optional[int], optional): top k documents to fetch. Defaults to 1.
        model_client (ModelClient): the model client to use.
        model_kwargs (Dict[str, Any], optional): the model kwargs. Defaults to {}.

    .. note::
        There is chance some queries might fail, which will lead to empty response None for that query in the List of RetrieverOutput. Users should handle this case.
    """

    def __init__(
        self,
        *,
        top_k: Optional[int] = 1,
        # the genearator kwargs
        model_client: ModelClient,
        model_kwargs: Dict[str, Any] = {},
        documents: Optional[RetrieverDocumentsType] = None,
        document_map_func: Optional[Callable[[Any], str]] = None,
    ):
        super().__init__()
        self.reset_index()
        self.generator = Generator(
            model_client=model_client,
            model_kwargs=model_kwargs,
            template=DEFAULT_LLM_AS_RETRIEVER_PROMPT_TEMPLATE,
            preset_prompt_kwargs={"top_k": top_k},
            output_processors=ListParser(),
        )

        self.top_k = top_k
        self.model_kwargs = model_kwargs

        if documents:
            self.build_index_from_documents(documents, document_map_func)

    def reset_index(self):
        self.indexed = False
        self.total_documents = 0

    def build_index_from_documents(
        self,
        documents: RetrieverDocumentsType,
        document_map_func: Optional[Callable[[Any], str]] = None,
    ):
        r"""prepare the user query input for the retriever"""
        if document_map_func:
            documents = [document_map_func(doc) for doc in documents]
        else:
            documents = documents
        self.total_documents = len(documents)
        self.generator.prompt.update_preset_prompt_kwargs(documents=documents)
        self.indexed = True

    def call(
        self,
        input: RetrieverStrQueriesType,
        top_k: Optional[int] = None,
        model_kwargs: Dict[str, Any] = {},
    ) -> RetrieverOutputType:
        """Retrieve the k relevant documents.

        Args:
            query_or_queries (RetrieverStrQueriesType): a string or a list of strings.
            top_k (Optional[int], optional): top k documents to fetch. Defaults to None.
            model_kwargs (Dict[str, Any], optional): the model kwargs.
             You can switch to another model provided by the same model client without reinitializing the retriever. Defaults to {}.

        Returns:
            RetrieverOutputType: the developers should be aware that the returned ``LLMRetrieverOutputType`` is actually a list of GeneratorOutput(:class:`GeneratorOutput <adalflow.core.types.GeneratorOutput>`), post processing is required depends on how you instruct the model to output in the prompt and what ``output_processors`` you set up.
            E.g. If the prompt is to output a list of indices and the ``output_processors`` is ``ListParser()``, then it return: GeneratorOutput(data=[indices], error=None, raw_response='[indices]')
        """
        assert self.indexed, "The retriever is not indexed yet."
        top_k = top_k or self.top_k
        queries = input if isinstance(input, list) else [input]
        retrieved_outputs: RetrieverOutputType = []

        for query in queries:
            prompt_kwargs = {
                "input_str": query,
                "top_k": top_k,
            }
            model_kwargs_to_use = self.model_kwargs.copy()
            model_kwargs_to_use.update(model_kwargs)
            response: GeneratorOutput = self.generator(
                prompt_kwargs=prompt_kwargs, model_kwargs=model_kwargs_to_use
            )
            if response.error or response.data is None:
                log.error(f"query: {query} failed to retrieve")
                log.error(f"error_message: {response.error}")
                log.error(f"raw_response: {response.raw_response}")
                log.error(f"response: {response.data}")
                retrieved_outputs.append(RetrieverOutput(doc_indices=[]))
                continue
            retrieved_outputs.append(
                RetrieverOutput(
                    doc_indices=response.data,
                    query=query,
                )
            )
        return retrieved_outputs

    def _extra_repr(self) -> str:
        s = f"top_k={self.top_k}, total_documents={self.total_documents},"
        return s



import dotenv
import os
import logging

log = logging.getLogger(__name__)


def setup_env(dotenv_path: str = ".env"):
    """Load environment variables from .env file."""

    if not os.path.exists(dotenv_path):
        raise FileNotFoundError(f"File not found: {dotenv_path}")

    try:
        dotenv.load_dotenv(dotenv_path=dotenv_path, verbose=True, override=False)
    except Exception as e:
        log.error(f"Error loading .env file: {e}")
        raise e



from .base_data_class import DataClass, required_field, DataClassFormatType

from .component import Component, DataComponent, func_to_data_component
from .container import Sequential, ComponentList
from .db import LocalDB
from .default_prompt_template import DEFAULT_ADALFLOW_SYSTEM_PROMPT
from .embedder import Embedder, BatchEmbedder
from .generator import Generator, BackwardEngine
from .model_client import ModelClient
from .string_parser import (
    YamlParser,
    JsonParser,
    IntParser,
    FloatParser,
    ListParser,
    BooleanParser,
)

from .prompt_builder import Prompt

from .retriever import Retriever
from .tokenizer import Tokenizer


from .types import (
    ModelType,
    ModelClientType,
    get_model_args,
    Embedding,
    Usage,
    TokenLogProb,
    EmbedderOutput,
    EmbedderInputType,
    EmbedderOutputType,
    BatchEmbedderInputType,
    BatchEmbedderOutputType,
    GeneratorOutput,
    GeneratorOutputType,
    Document,
    RetrieverQueryType,
    RetrieverStrQueryType,
    RetrieverQueriesType,
    RetrieverStrQueriesType,
    RetrieverDocumentType,
    RetrieverStrDocumentType,
    RetrieverDocumentsType,
    RetrieverOutput,
    RetrieverOutputType,
    UserQuery,
    AssistantResponse,
    DialogTurn,
    Conversation,
)
from adalflow.utils.registry import EntityMapping

__all__ = [
    "LocalDB",
    "Component",
    "DataComponent",
    "func_to_data_component",
    "Sequential",
    "ComponentList",
    "DataClass",
    "DataClassFormatType",
    "required_field",
    "Generator",
    "BackwardEngine",
    "Prompt",
    "DEFAULT_ADALFLOW_SYSTEM_PROMPT",
    # "Parameter",
    "required_field",
    "ModelClient",
    "Embedder",
    "BatchEmbedder",
    "Retriever",
    "GeneratorOutput",
    "GeneratorOutputType",
    "ModelType",
    "ModelClientType",
    "get_model_args",
    "Embedding",
    "Usage",
    "TokenLogProb",
    "EmbedderOutput",
    "EmbedderInputType",
    "EmbedderOutputType",
    "BatchEmbedderInputType",
    "BatchEmbedderOutputType",
    "Document",
    "RetrieverQueryType",
    "RetrieverStrQueryType",
    "RetrieverQueriesType",
    "RetrieverStrQueriesType",
    "RetrieverDocumentType",
    "RetrieverStrDocumentType",
    "RetrieverDocumentsType",
    "RetrieverOutput",
    "RetrieverOutputType",
    "UserQuery",
    "AssistantResponse",
    "DialogTurn",
    "Conversation",
    "Tokenizer",
    # Parsers
    "YamlParser",
    "JsonParser",
    "IntParser",
    "FloatParser",
    "ListParser",
    "BooleanParser",
]

for name in __all__:
    EntityMapping.register(name, globals()[name])



"""Default Dataset, DataLoader similar to `utils.data` in PyTorch.

You can also use those provided by PyTorch or huggingface/datasets."""

from typing import Union, Tuple, List, Sequence, TypeVar, Generic
import numpy as np
import random

T_co = TypeVar("T_co", covariant=True)


# TODO: consider directly use torch.utils.data in the future
class Dataset(Generic[T_co]):
    r"""An abstract class representing a :class:`Dataset`.

    All datasets that represent a map from keys to data samples should subclass
    it. All subclasses should overwrite :meth:`__getitem__`, supporting fetching a
    data sample for a given key. Subclasses could also optionally overwrite
    :meth:`__len__`, which is expected to return the size of the dataset by many
    :class:`~torch.utils.data.Sampler` implementations and the default options
    of :class:`~torch.utils.data.DataLoader`. Subclasses could also
    optionally implement :meth:`__getitems__`, for speedup batched samples
    loading. This method accepts list of indices of samples of batch and returns
    list of samples.

    .. note::
      :class:`~torch.utils.data.DataLoader` by default constructs an index
      sampler that yields integral indices.  To make it work with a map-style
      dataset with non-integral indices/keys, a custom sampler must be provided.
    """

    def __getitem__(self, index) -> T_co:
        raise NotImplementedError("Subclasses of Dataset should implement __getitem__.")

    def __len__(self) -> int:
        raise NotImplementedError("Subclasses of Dataset should implement __len__.")


class Subset(Dataset[T_co]):
    r"""
    Subset of a dataset at specified indices.

    Args:
        dataset (Dataset): The whole Dataset
        indices (sequence): Indices in the whole set selected for subset
    """

    dataset: Dataset[T_co]
    indices: Sequence[int]

    def __init__(self, dataset: Dataset[T_co], indices: Sequence[int]) -> None:
        self.dataset = dataset
        self.indices = indices

    def __getitem__(self, idx):
        if isinstance(idx, list):
            return self.dataset[[self.indices[i] for i in idx]]
        return self.dataset[self.indices[idx]]

    def __getitems__(self, indices: List[int]) -> List[T_co]:
        # add batched sampling support when parent dataset supports it.
        # see torch.utils.data._utils.fetch._MapDatasetFetcher
        if callable(getattr(self.dataset, "__getitems__", None)):
            return self.dataset.__getitems__([self.indices[idx] for idx in indices])  # type: ignore[attr-defined]
        else:
            return [self.dataset[self.indices[idx]] for idx in indices]

    def __len__(self):
        return len(self.indices)


class DataLoader:
    __doc__ = r"""A simplified version of PyTorch DataLoader.

    The biggest difference is not to handle tensors, but to handle any type of data."""

    def __init__(
        self, dataset, batch_size: int = 4, shuffle: bool = True, seed: int = 42
    ):
        self.dataset = dataset
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.seed = seed

        self.indices = np.arange(len(dataset))
        # if self.shuffle:
        #     np.random.shuffle(self.indices)
        self.current_index = 0
        self.max_steps = self.__len__()
        self.step_index = 0

    def set_max_steps(self, max_steps: int):
        self.max_steps = max_steps

    def __iter__(self):
        if self.shuffle:
            if self.seed is not None:
                np.random.seed(self.seed)  # Use the provided seed
            np.random.shuffle(self.indices)
        self.current_index = 0
        return self

    def __len__(self):
        return (len(self.dataset) + self.batch_size - 1) // self.batch_size

    def __next__(self) -> Union[np.ndarray, Tuple]:
        # if self.current_index >= len(self.dataset):
        #     raise StopIteration

        if self.current_index >= len(self.dataset):
            if self.shuffle:
                if self.seed is not None:
                    np.random.seed(self.seed)  # Use the same seed for reshuffle
                np.random.shuffle(self.indices)  # Reshuffle for the new epoch
            self.current_index = 0
            if self.step_index < self.max_steps:
                pass
            else:
                raise StopIteration
            # raise StopIteration

        batch_indices = self.indices[
            self.current_index : self.current_index + self.batch_size
        ]
        batch_data = [self.dataset[int(i)] for i in batch_indices]

        if isinstance(batch_data[0], tuple):
            batch_data = tuple(zip(*batch_data))
        else:
            batch_data = np.array(batch_data)

        self.current_index += self.batch_size
        self.step_index += 1

        return batch_data


def subset_dataset(dataset, num_samples: int):
    r"""This function will be useful for testing and debugging purposes."""
    num_samples = min(num_samples, len(dataset))
    random_subset_indices = random.sample(range(len(dataset)), num_samples)
    return Subset(dataset, random_subset_indices)



"""Implementation of TextGrad: Automatic “Differentiation” via Text.
This code is not used as we treat the non-optimizable version of LLM judge as a form of eval_fn.
We use class EvalFnToTextLoss instead as of today 12/9/2024"""

from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:
    from adalflow.core import ModelClient
    from ..parameter import Parameter


from adalflow.optim.loss_component import LossComponent
from typing import Dict
from copy import deepcopy
import logging

log = logging.getLogger(__name__)


TEXT_LOSS_TEMPLATE = r"""<START_OF_SYSTEM_PROMPT>
{{eval_system_prompt}}
<END_OF_SYSTEM_PROMPT>
<USER>
{{eval_user_prompt}}
</USER>
"""


class LLMAsTextLoss(LossComponent):
    __doc__ = r"""Evaluate the final RAG response using an LLM judge.

    The LLM judge will have:
    - eval_system_prompt: The system prompt to evaluate the response.
    - y_hat: The response to evaluate.
    - Optional: y: The correct response to compare against.

    The loss will be a Parameter with the evaluation result and can be used to compute gradients.
    This loss use LLM/Generator as the computation/transformation operator, so it's gradient will be
    found from the Generator's backward method.
    """

    def __init__(
        self,
        prompt_kwargs: Dict[str, Union[str, "Parameter"]],
        model_client: "ModelClient",
        model_kwargs: Dict[str, object],
    ):
        from adalflow.core.generator import Generator
        from adalflow.optim.parameter import Parameter

        super().__init__()
        prompt_kwargs = deepcopy(prompt_kwargs)
        # TODO: do we really need to convert str to Parameter? what if not?
        for key, value in prompt_kwargs.items():
            if isinstance(value, str):
                prompt_kwargs[key] = Parameter(
                    data=value, requires_opt=False, role_desc=key
                )
        self.prompt_kwargs = prompt_kwargs
        # this is llm as judge (loss) to get the loss
        self.loss_llm = Generator(
            name="llm_judge",
            model_client=model_client,
            model_kwargs=model_kwargs,
            template=TEXT_LOSS_TEMPLATE,
            prompt_kwargs=prompt_kwargs,
        )

    # def __call__(self, *args, **kwargs):
    #     return self.forward(*args, **kwargs)

    def forward(self, *args, **kwargs) -> "Parameter":

        return self.loss_llm.forward(*args, **kwargs)



__version__ = "1.0.0.beta.3"

from adalflow.core.component import (
    Component,
    DataComponent,
    FuncComponent,
    FuncDataComponent,
    func_to_data_component,
    func_to_component,
)
from adalflow.core.container import Sequential, ComponentList
from adalflow.core.base_data_class import DataClass, DataClassFormatType, required_field

from adalflow.optim.grad_component import GradComponent
from adalflow.core.generator import Generator

from adalflow.core.types import (
    GeneratorOutput,
    EmbedderOutput,
    RetrieverOutput,
    Document,
)
from adalflow.core.model_client import ModelClient
from adalflow.core.embedder import Embedder

# parser
from adalflow.core.string_parser import (
    YamlParser,
    JsonParser,
    IntParser,
    FloatParser,
    ListParser,
    BooleanParser,
)
from adalflow.core.retriever import Retriever
from adalflow.components.output_parsers import (
    YamlOutputParser,
    JsonOutputParser,
    ListOutputParser,
)
from adalflow.components.output_parsers.dataclass_parser import DataClassParser

from adalflow.core.prompt_builder import Prompt

# optimization
from adalflow.optim import (
    Optimizer,
    DemoOptimizer,
    TextOptimizer,
    Parameter,
    AdalComponent,
    Trainer,
    BootstrapFewShot,
    TGDOptimizer,
    EvalFnToTextLoss,
    LLMAsTextLoss,
)

from adalflow.optim.types import ParameterType
from adalflow.utils import setup_env, get_logger

from adalflow.components.model_client import (
    OpenAIClient,
    GoogleGenAIClient,
    GroqAPIClient,
    OllamaClient,
    TransformersClient,
    AnthropicAPIClient,
    CohereAPIClient,
    BedrockAPIClient,
)

# data pipeline
from adalflow.components.data_process.text_splitter import TextSplitter
from adalflow.components.data_process.data_components import ToEmbeddings

__all__ = [
    "Component",
    "DataComponent",
    "FuncComponent",
    "FuncDataComponent",
    "func_to_data_component",
    "func_to_component",
    # dataclass
    "DataClass",
    "DataClassFormatType",
    "required_field",
    # Container
    "Sequential",
    "ComponentList",
    # Grad Component
    "GradComponent",
    # Functional Component
    "ModelClient",
    "Generator",
    "Embedder",
    "Retriever",
    "Parameter",
    "AdalComponent",
    "Trainer",
    "BootstrapFewShot",
    "TGDOptimizer",
    "EvalFnToTextLoss",
    "LLMAsTextLoss",
    "setup_env",
    "get_logger",
    "Prompt",
    # Parsers
    "YamlParser",
    "JsonParser",
    "IntParser",
    "FloatParser",
    "ListParser",
    "BooleanParser",
    "Parser",
    "FuncParser",
    # Output Parsers with dataclass formatting
    "YamlOutputParser",
    "JsonOutputParser",
    "ListOutputParser",
    "DataClassParser",
    # Data Pipeline
    "TextSplitter",
    "ToEmbeddings",
    # Types
    "GeneratorOutput",
    "EmbedderOutput",
    "RetrieverOutput",
    "Document",
    # Optimizer types
    "Optimizer",
    "DemoOptimizer",
    "TextOptimizer",
    # parameter types
    "ParameterType",
    # model clients
    "OpenAIClient",
    "GoogleGenAIClient",
    "GroqAPIClient",
    "OllamaClient",
    "TransformersClient",
    "AnthropicAPIClient",
    "CohereAPIClient",
    "BedrockAPIClient",
]



"""Text-grad optimizer and prompts. Also combined methods from ORPO llm optimizer.

With the auto-diff gradients, it made it possible to optimize any prompt parameter in a task pipeline.

https://arxiv.org/abs/2309.03409
Source code: https://github.com/google-deepmind/opro
"""

from typing import List, Dict, TYPE_CHECKING, Optional, Any
import logging
import re
from dataclasses import field, dataclass

from adalflow.optim.optimizer import TextOptimizer, ParamsT
from adalflow.optim.text_grad.backend_engine_prompt import VARIABLE_AND_PEERS_INFO
from adalflow.optim.parameter import Parameter

from adalflow.core.base_data_class import DataClass
from adalflow.utils.logger import printc
from adalflow.core.types import GeneratorOutput


if TYPE_CHECKING:
    from adalflow.core import ModelClient


log = logging.getLogger(__name__)


@dataclass
class HistoryPrompt(DataClass):
    id: str
    value: str
    eval_score: float
    method: str = field(default=None)
    reasoning: str = field(default=None)


####################################################################################################
# Textual Gradient Descent Optimizer
####################################################################################################
# TODO: make the uesr message to task instruction

TEXT_GRAD_DESC_TEMPLATE = r"""<START_OF_SYSTEM_PROMPT>
{{optimizer_system_prompt}}
<END_OF_SYSTEM_PROMPT>
<START_OF_USER_MESSAGE>
You are {{steps}} steps since your last improvement.
Update the value more rapidly when steps are larger than 3.
{# Variable and peers info #}
<START_OF_VARIABLE_AND_PEERS_INFO>
{{variable_and_peers_info}}
<END_OF_VARIABLE_AND_PEERS_INFO>
{# system trainable variables #}
{% if system_variables %}
<START_OF_SYSTEM_VARIABLES>
The target variable is used together with these system variables besides of its peers:
{% for system_variable in system_variables %}
{{loop.index}}.
Name: {{system_variable.name}}
Type: {{system_variable.param_type}}
Description: {{system_variable.role_desc}}
WILL_BE_OPTIMIZED: {{system_variable.requires_opt}}
Value: {{system_variable.prompt_data}}
{% endfor %}
Strategically plan the role of each system variable to collaborate with each other for final correct answer.
<END_OF_SYSTEM_VARIABLES>
{% endif %}
{# OPRO past history #}
{% if past_history %}
<START_OF_HISTORY_PERFORMANCE>
Here are the best past iterations.
{% for history in past_history %}
{{loop.index}}. {{history}}
{% endfor %}
IMPORTANT: Your goal is to generate new variable that score higher than all past iterations.
<END_OF_HISTORY_PERFORMANCE>
{% endif %}
{# Multi-proposal history #}
{% if failed_proposals %}
<START_OF_CURRENT_ITERATION>
same batch, same feedback: Here are the values you have tried that have not improved the score.(scored <= {{best_score}}):
{% for failed_proposal in failed_proposals %}
{{loop.index}}. {{failed_proposal}}
{% endfor %}
You MUST approach differently from the above methods.
<END_OF_CURRENT_ITERATION>
{% endif %}
{# Feedback #}
{% if variable_grad %}
<START_OF_CONTEXT_FEEDBACK>
Here are the context and feedback for the variable:
{{variable_grad}}
<END_OF_CONTEXT_FEEDBACK>
{% endif %}
{# Constraints #}
{% if constraint_text %}
You must follow the following constraints:
<CONSTRAINTS>{{constraint_text}}</CONSTRAINTS>
{% endif %}
{# In-context examples #}
{% if in_context_examples %}
You must base on the following examples when modifying the {{variable_desc}}:
<EXAMPLES>{{in_context_examples}}</EXAMPLES>
{% endif %}
<END_OF_USER_MESSAGE>
"""
# NO OPRO history
# TEXT_GRAD_DESC_TEMPLATE = r"""<START_OF_SYSTEM_PROMPT>
# {{optimizer_system_prompt}}
# <END_OF_SYSTEM_PROMPT>
# <START_OF_USER_MESSAGE>
# You are {{steps}} steps since your last improvement.
# Update the value more rapidly when steps are larger than 3.
# {# Variable and peers info #}
# <START_OF_VARIABLE_AND_PEERS_INFO>
# {{variable_and_peers_info}}
# <END_OF_VARIABLE_AND_PEERS_INFO>
# {# system trainable variables #}
# {% if system_variables %}
# <START_OF_SYSTEM_VARIABLES>
# The target variable is used together with these system variables besides of its peers:
# {% for system_variable in system_variables %}
# {{loop.index}}.
# Name: {{system_variable.name}}
# Type: {{system_variable.param_type}}
# Description: {{system_variable.role_desc}}
# WILL_BE_OPTIMIZED: {{system_variable.requires_opt}}
# Vaule: {{system_variable.prompt_data}}
# {% endfor %}
# Strategically plan the role of each system variable to collaborate with each other for final correct answer.
# <END_OF_SYSTEM_VARIABLES>
# {% endif %}
# {# Feedback #}
# {% if variable_grad %}
# <START_OF_CONTEXT_FEEDBACK>
# Here are the context and feedback for the variable:
# {{variable_grad}}
# <END_OF_CONTEXT_FEEDBACK>
# {% endif %}
# {# Constraints #}
# {% if constraint_text %}
# You must follow the following constraints:
# <CONSTRAINTS>{{constraint_text}}</CONSTRAINTS>
# {% endif %}
# {# In-context examples #}
# {% if in_context_examples %}
# You must base on the following examples when modifying the {{variable_desc}}:
# <EXAMPLES>{{in_context_examples}}</EXAMPLES>
# {% endif %}
# <END_OF_USER_MESSAGE>
# """
# NO promt engineering techniques
# OPTIMIZER_SYSTEM_PROMPT = r"""You are an excellent prompt engineer tasked with instruction and demonstration tuning a compound LLM system.
# Your task is to refine a variable/prompt based on feedback from a batch of input data points.

# The variable is either input or output of a functional component where the component schema will be provided.
# If the same DataID has multiple gradients, it means this component/variable is called multiple times in the compound system(with a cycle) in the same order as it appears in the gradient list.

# You Must edit the current variable with one of the following editing methods.
# You can not rewrite everything all at once:

# You have Four Editing Methods:
# 1. ADD new elements(instruction) to address each specific feedback.
# 2. ADD Examples (e.g., input-reasoning-answer) for tasks that require strong reasoning skills.
# 3. Rephrase existing instruction(for more clarity), Replace existing sample with another, to address the feedback.
# 4. DELETE unnecessary words to improve clarity.

# Your final action/reasoning  = one of FOUR editing method

# You must stick to these instructions:
# 1. **MUST Resolve concerns raised in the feedback** while preserving the positive aspects of the original variable.
# 2. **Observe past performance patterns** to retain good qualities in the variable and past failed ones to try things differently.
# 3. **System Awareness**: When other system variables are given, ensure you understand how this variable works in the whole system.
# 4. **Peer Awareness**: This variable works together with Peer variables, ensure you are aware of their roles and constraints.
# 5. **Batch Awareness**: You are optimizing a batch of input data, ensure the change applys to the whole batch (except while using demonstration.)

# {{output_format_str}}

# {% if instruction_to_optimizer %}
# **Additional User Instructions**: {{instruction_to_optimizer}}
# {% endif %}
# """

OPTIMIZER_SYSTEM_PROMPT = r"""You are an excellent prompt engineer tasked with instruction and demonstration tuning a compound LLM system.
Your task is to refine a variable/prompt based on feedback from a batch of input data points.

The variable is either input or output of a functional component where the component schema will be provided.
If the same DataID has multiple gradients, it means this component/variable is called multiple times in the compound system(with a cycle) in the same order as it appears in the gradient list.

You Must edit the current variable with one of the following editing methods.
You can not rewrite everything all at once:

You have Four Editing Methods:
1. ADD new elements(instruction) to address each specific feedback.
2. ADD Examples (e.g., input-reasoning-answer) for tasks that require strong reasoning skills.
3. Rephrase existing instruction(for more clarity), Replace existing sample with another, to address the feedback.
4. DELETE unnecessary words to improve clarity.

These SIX prompting techniques can be a helpful direction.
1. Set Context and Role: Establish a specific identity or domain expertise for the AI to guide style, knowledge, and constraints.
2. Be Specific, Clear, and Grammarly correct: Clearly define instructions, desired format, and constraints to ensure accurate and relevant outputs with regards to the feedback.
3. Illicit reasoning: "chain-of-thought" (e.g. "think step by step") helps the model reason better.
4. Examples: Construct examples(e.g., input(optional)-reasoning(required)-answer) especially for tasks that require strong reasoning skills.
5. Leverage Constraints and Formatting: Explicitly direct how the answer should be structured (e.g., bullet points, tables, or tone).
6. Self-Consistency / Verification Prompts: Prompt the model to check its own logic for errors, inconsistencies, or missing details.

Your final action/reasoning  = one of FOUR editing method + one of SIX prompting technique.

You must stick to these instructions:
1. **MUST Resolve concerns raised in the feedback** while preserving the positive aspects of the original variable.
2. **Observe past performance patterns** to retain good qualities in the variable and past failed ones to try things differently.
3. **System Awareness**: When other system variables are given, ensure you understand how this variable works in the whole system.
4. **Peer Awareness**: This variable works together with Peer variables, ensure you are aware of their roles and constraints.
5. **Batch Awareness**: You are optimizing a batch of input data, ensure the change applys to the whole batch (except while using demonstration.)

{{output_format_str}}

{% if instruction_to_optimizer %}
**Additional User Instructions**: {{instruction_to_optimizer}}
{% endif %}
"""


@dataclass
class Instruction(DataClass):
    __doc__ = "Structure variable values for instructions. Can be used in the history of instructions."
    text: str = field(metadata={"desc": "The instruction text"})
    score: float = field(
        metadata={"desc": "The score of the instruction, range from 0 to 1"}
    )
    responses: Optional[List[str]] = field(
        metadata={"desc": "The responses of using the instruction"}, default=None
    )
    gts: Optional[List[str]] = field(
        metadata={"desc": "The ground truth of the task"}, default=None
    )


@dataclass
class TGDData(DataClass):
    reasoning: str = field(metadata={"desc": "Why the variable is proposed this way"})
    method: str = field(
        metadata={
            "desc": "The final method used to propose the variable (prompting + editing)"
        },
    )

    proposed_variable: str = field(
        metadata={"desc": "The proposed variable"},
        default=None,
    )


@dataclass
class TGDOptimizerTrace(DataClass):
    api_kwargs: Dict[str, Any] = field(
        metadata={
            "desc": "The api_kwargs for components like Generator and Retriever that pass to the model client"
        },
        default=None,
    )
    output: TGDData = field(
        metadata={"desc": "The output of the TGD optimizer"}, default=None
    )


new_variable_tags = ["<VARIABLE>", "</VARIABLE>"]


def extract_new_variable(text: str) -> str:
    pattern = re.compile(r"<VARIABLE>(.*?)</VARIABLE>", re.DOTALL)

    # Find all matches
    matches = pattern.findall(text)

    if len(matches) == 0:
        return text.strip()
    log.debug(f"Extracted new variable: {matches[0].strip()}")
    return matches[0].strip()


class TGDOptimizer(TextOptimizer):
    __doc__ = """Textual Gradient Descent(LLM) optimizer for text-based variables."""

    proposing: bool = False
    params: ParamsT
    constraints: List[str]
    params_history: Dict[str, List[HistoryPrompt]] = {}  # id to history
    failed_proposals: Dict[str, List[HistoryPrompt]] = {}  # only need the value
    current_tgd_output: Dict[str, Optional[TGDData]] = (
        {}
    )  # id to output, hold all of the data
    one_parameter_at_a_time: bool

    def __init__(
        self,
        params: ParamsT,
        model_client: "ModelClient",
        model_kwargs: Dict[str, object] = {},
        constraints: List[str] = None,
        optimizer_system_prompt: str = OPTIMIZER_SYSTEM_PROMPT,
        in_context_examples: List[str] = None,  # TODO: in-context examples
        max_past_history: int = 3,
        max_failed_proposals: int = 5,  # quite effective
        steps_from_last_improvement: int = 0,
        one_parameter_at_a_time: bool = True,
    ):
        from adalflow.core.generator import Generator
        from adalflow.core import Prompt
        from adalflow.components.output_parsers.dataclass_parser import DataClassParser

        super().__init__()
        self.params = params  # all parameters of prompts (even if non-trainable)

        self.constraints = constraints or []
        self.data_class = TGDData
        self.output_parser = DataClassParser(
            data_class=self.data_class, return_data_class=True, format_type="json"
        )
        self.optimizer_system_prompt = Prompt(
            template=optimizer_system_prompt,
            prompt_kwargs={
                "output_format_str": self.output_parser.get_output_format_str(),
            },
        )
        self.variable_and_peers_info = Prompt(
            template=VARIABLE_AND_PEERS_INFO,
        )
        self.do_constrained = len(self.constraints) > 0
        # self.new_variable_tags = new_variable_tags
        self.in_context_examples = in_context_examples or []
        self.do_in_context_examples = len(self.in_context_examples) > 0

        self.llm_optimizer = Generator(
            model_client=model_client,
            model_kwargs=model_kwargs,
            template=TEXT_GRAD_DESC_TEMPLATE,
            output_processors=self.output_parser,
        )

        self.max_past_history = max_past_history
        self.max_failed_proposals = max_failed_proposals
        self.steps_from_last_improvement = steps_from_last_improvement

        self.target_param_index = None
        self.one_parameter_at_a_time = False  # one_parameter_at_a_time
        # initate the past history for each parameter
        for param in self.params:
            self.params_history[param.id] = []
            self.failed_proposals[param.id] = []
            self.current_tgd_output[param.id] = None

    @property
    def constraint_text(self):
        """
        Returns a formatted string representation of the constraints.

        :return: A string containing the constraints in the format "Constraint {index}: {constraint}".
        :rtype: str
        """
        constraints_ordered = [
            f"Constraint {i+1}: {constraint}"
            for i, constraint in enumerate(self.constraints)
        ]
        return "\n".join(constraints_ordered)

    def increment_steps_from_last_improvement(self):
        self.steps_from_last_improvement += 1

    def reset_steps_from_last_improvement(self):
        self.steps_from_last_improvement = 0

    def add_score_to_params(self, val_score: float):
        for param in self.params:
            self.add_score_to_current_param(param.id, param, val_score)

    def add_score_to_current_param(self, param_id: str, param: Parameter, score: float):
        if param_id not in self.params_history:
            raise ValueError(f"Parameter {param_id} not found in the history.")
        if param.id != param_id:
            raise ValueError(
                f"Parameter {param_id} does not match the target parameter."
            )

        history = HistoryPrompt(
            id=param_id,
            value=str(param.data),
            eval_score=score,
        )
        self.add_history(param_id, history)

    def add_history(self, param_id: str, history: HistoryPrompt):
        if param_id not in self.params_history:
            self.params_history[param_id] = []
        # check if the value is already in the history, if so, replace it with the new one
        for i, h in enumerate(self.params_history[param_id]):
            if h.value == history.value:
                self.params_history[param_id].pop(i)
                break
        self.params_history[param_id].append(history)
        # sort score from the highest to the lowest
        self.params_history[param_id] = sorted(
            self.params_history[param_id], key=lambda x: x.eval_score, reverse=True
        )

        # delete the lowest score if it exceeds the max_past
        if len(self.params_history[param_id]) > self.max_past_history:
            for _ in range(len(self.params_history[param_id]) - self.max_past_history):
                self.params_history[param_id].pop()

    def render_history(self, param_id: str) -> List[str]:
        if param_id not in self.params_history:
            return []

        return [
            history.to_yaml(exclude=["id", "method", "reasoning"])
            for history in self.params_history[param_id]
        ]

    def add_failed_proposal(self):
        """Save a copy of the current value of the parameter in the failed proposals."""
        for param in self.params:
            current_tgd_output = self.current_tgd_output.get(param.id, None)
            failed_proposal = HistoryPrompt(
                id=param.id,
                value=param.data,
                eval_score=None,
                method=(current_tgd_output.method if current_tgd_output else None),
                reasoning=(
                    current_tgd_output.reasoning if current_tgd_output else None
                ),
            )
            self.failed_proposals[param.id].append(failed_proposal)
            if len(self.failed_proposals[param.id]) > self.max_failed_proposals:
                for _ in range(
                    len(self.failed_proposals[param.id]) - self.max_failed_proposals
                ):
                    self.failed_proposals[param.id].pop(0)

    def render_failed_proposals(self, param_id: str) -> List[str]:
        if param_id not in self.failed_proposals:
            return []
        return [
            history.to_yaml(exclude=["id", "eval_score", "value"])
            for history in self.failed_proposals[param_id]
        ]

    def _get_user_prompt_kwargs(self, param: Parameter) -> Dict[str, str]:

        system_params = [
            p.get_param_info()
            for p in self.params
            if p.id != param.id and p not in param.peers
        ]
        printc(f"system_params: {system_params}", color="blue")
        peers_params = [p.get_param_info() for p in param.peers]
        variable_and_peer_info = self.variable_and_peers_info.call(
            variable=param.get_param_info(), peers=peers_params
        )

        variable_grad = param.get_gradients_component_schema(skip_correct_sample=False)

        user_prompt_kwargs = {
            "variable_and_peers_info": variable_and_peer_info,
            "variable_grad": variable_grad,
            # constraints
            "constraint_text": self.constraint_text if self.do_constrained else None,
            # in-context examples
            "in_context_examples": (
                "\n".join(self.in_context_examples)
                if self.do_in_context_examples
                else None
            ),
            # past history
            "past_history": (
                self.render_history(param.id) if self.max_past_history else None
            ),
            # failed proposals
            "failed_proposals": (
                self.render_failed_proposals(param.id)
                if self.max_failed_proposals
                else None
            ),
            "best_score": (
                self.params_history[param.id][0].eval_score
                if self.params_history[param.id]
                else "N/A"
            ),
            "system_variables": system_params,
            "steps": self.steps_from_last_improvement,
        }

        return user_prompt_kwargs

    # TODO: better way to update the gradient memory
    def update_gradient_memory(self, param: Parameter):
        self.gradient_memory_dict[param.id].append(
            {"value": param.get_gradient_and_context_text(skip_correct_sample=True)}
        )

    def zero_grad(self):
        for p in self.params:
            p.reset_gradients()

        # reset the failded proposals
        for param in self.params:
            self.failed_proposals[param.id] = []

    def set_target_param(self):
        # iterate through all indexes in cycle
        if self.target_param_index is None:
            self.target_param_index = 0
        else:
            self.target_param_index = (self.target_param_index + 1) % len(self.params)

    # TODO: in the future can propose multiple values at once
    def propose(self):
        r"""Proposing a value while keeping previous value saved on parameter."""
        if self.proposing:
            raise ValueError("Already proposing a value.")

        printc("Proposing a new value.", color="magenta")

        # no cache so that new proposal can be made
        no_cache = True
        # print("Proposing a new value.")

        for idx, param in enumerate(self.params):
            if not param.requires_opt:
                log.info(
                    f"Skipping {param.role_desc} as it does not require optimization."
                )
                continue
            if self.one_parameter_at_a_time and idx != self.target_param_index:
                continue

            system_prompt = self.optimizer_system_prompt(
                param_type=str(param.param_type),
                instruction_to_optimizer=param.instruction_to_optimizer,
            )
            user_prompt_kwargs = self._get_user_prompt_kwargs(param)
            prompt_kwargs = {
                "optimizer_system_prompt": system_prompt,
                **user_prompt_kwargs,
            }
            # turn off cache
            try:
                response: GeneratorOutput = self.llm_optimizer.call(
                    prompt_kwargs=prompt_kwargs, use_cache=not no_cache
                )
            except Exception as e:
                printc(f"Error in the optimizer: {e}", color="red")
                raise e
            if not isinstance(response, GeneratorOutput):
                raise TypeError(f"Wrong response type: {type(response)}")

            prompt_str = self.llm_optimizer.get_prompt(**prompt_kwargs)
            log.debug(f"TGD LLM optimizer prompt: {prompt_str}")
            printc(f"TGD LLM optimizer prompt: {prompt_str}", color="blue")
            proposed_data: TGDData = (
                response.data
                if response.data is not None
                else TGDData(
                    reasoning="No reasoning",
                    proposed_variable=response.raw_response,
                    method="No method",
                )
            )
            # save current tgd output data
            self.current_tgd_output[param.id] = proposed_data
            printc(f"Response from the optimizer: {response}", color="blue")

            log.info(f"Response from the optimizer: {response}")
            # if not proposed_data.update:
            #     printc(f"No update is required for {param.name}", color="yellow")
            #     param.propose_data(param.data)
            # else:  # TODO: should always trace the initial data
            improved_variable = proposed_data.proposed_variable
            if (
                improved_variable
                and improved_variable != param.data
                and improved_variable != ""
            ):
                param.propose_data(improved_variable)
            else:
                param.propose_data(param.data)
            param.trace_optimizer(api_kwargs=prompt_str, response=response)

        self.proposing = True

    def revert(self):
        """Revert to the previous value when the evaluation is worse."""
        if not self.proposing:
            raise ValueError("Not proposing a value.")
        for idx, param in enumerate(self.params):
            if not param.requires_opt:
                continue
            if self.one_parameter_at_a_time and idx != self.target_param_index:
                continue
            param.revert_data()
            param.trace_optimizer(api_kwargs=None, response=None)
        self.proposing = False

    def step(self):
        """Discard the previous value and keep the proposed value."""
        if not self.proposing:
            raise ValueError("Not proposing a value.")
        for idx, param in enumerate(self.params):
            if not param.requires_opt:
                continue
            if self.one_parameter_at_a_time and idx != self.target_param_index:
                continue

            param.step_data()

        self.proposing = False

    def to_dict(self):
        return {
            "template": TEXT_GRAD_DESC_TEMPLATE,
            "optimizer_system_prompt": OPTIMIZER_SYSTEM_PROMPT,
            "VARIABLE_AND_PEERS_INFO": VARIABLE_AND_PEERS_INFO,
            "params": self.params,
            "constraints": self.constraints,
            "params_history": self.params_history,
            "failed_proposals": self.failed_proposals,
            "max_past_history": self.max_past_history,
            "max_failed_proposals": self.max_failed_proposals,
            "steps_from_last_improvement": self.steps_from_last_improvement,
        }


if __name__ == "__main__":
    # test the prompt history
    data = {
        "id": "1",
        "value": "test",
        "eval_score": 0.5,
    }
    history = HistoryPrompt.from_dict(data)

    print(history)
    history_yaml = history.to_yaml()
    print(history_yaml)

    template = r"""<START_OF_SYSTEM_PROMPT>
    {% if past_history %}
    Here are the past iterations of this variable along with the validation score.
    {% for history in past_history %}
    {{loop.index}}. {{history}}
    {% endfor %}
    {% endif %}
    <END_OF_SYSTEM_PROMPT>"""

    import adalflow as adal

    prompt = adal.Prompt(template=template)

    histories = [history_yaml]
    prompt_kwargs = {
        "past_history": histories,
    }
    response = prompt(**prompt_kwargs)
    # print(response)
